//  Brian Marshall      bmarshal@agt.net
//......................................

//  Employee functions.

#include <stdlib.h>
#include <string.h>
#include "workforc.h"

//---------------------------------------------------------
//  Setup the dynamic components at problem-start time.

void  employee::problem_setup (char *rec)
{
  if (badge != Badge_Id) abort ("Emp-ps1");

  if ( ! (day = new empl_day[Num_Days]) ) {
    wlog ("Error getting memory for employee-days.\n");
    wfexit(1);
  }

  rec_setup (rec);

  i_left_bhnd = Trivial;
  i_been_prob = Trivial;

  i_strategic = iemp - ((iemp / 3) * 3);    // 0, 1, 2.

  cur_day = day + Week_Start;

  g = Ok;
  g_prev = Imposible;
  g_bsf = Imposible;

  num_fail = 0;

  days_on_need  = days_per_week;
  days_off_need = off_per_week;

  done = days_per_week == 0;

  min_ttp = 0;      // Minutes to this point.

  short int iday;
  for (iday=0; iday<Num_Days; iday++) {
    empl_day *ed = day + iday;

    ed->badge = Badge_Id;
    ed->iday = iday;
    if (iday == Num_Days-1) {
      ed->next_day = null;
    } else {
      ed->next_day = ed + 1;
    }
    if (iday == 0) {
      ed->prev_day = null;
    } else {
      ed->prev_day = ed - 1;
    }
    ed->emp = this;

    ed->problem_setup();

  }    // End of loop through days.

  int len = strlen (id);
  if (len < 1 || len > 12) abort ("Emp-1");
  len = strlen (name);
  if (len < 1 || len > 20) abort ("Emp-2");
  if (days_per_week < 0 || days_per_week > 7) abort ("Emp-3");
  if (off_per_week != Week_Len-days_per_week) abort ("Emp-4");
  if (seniority < 1 || seniority > 99) abort ("Emp-9");
  if (cur_day->iday != Week_Start) abort ("Emp-13");

}

//---------------------------------------------------------
//  Apply the information in the record from the employee file.

void  employee::rec_setup (char *rec)
{
  if (badge != Badge_Id) abort ("Emp-rs1");

  char  buf[80];

  istrstream istr(rec,strlen(rec));     // Make a string-stream.
  if (!istr) {
    wlog ("Error creating strstream from record: \n");
    wlog2 (rec, "\n  ABORT.\n");
    wfexit(1);
  }

  strcpy (id, get_s(&istr,12));
  strcpy (name, get_s(&istr,20));

  strcpy (buf, get_s(&istr,20));
  days_per_week  = atoi(buf);
  if (days_per_week < 1 || days_per_week > 7) {
    wlog3 ("Incorrect Days-per-Week  \"", buf, "\" on line: \n");
    wlog2 (rec, "\n  (must be between 1 and 7).  ABORT.\n");
    wfexit(1);
  }

  off_per_week = Week_Len - days_per_week;

  strcpy (buf, get_s(&istr,20));
  min_per_week = hr_minutes (buf);
  if (min_per_week < 0 || min_per_week > 6000) {
    wlog3 ("Incorrect Hours-per-Week  \"", buf, "\" on line: \n");
    wlog2 (rec, "\n  ABORT.\n");
    wfexit(1);
  }

  strcpy (buf, get_s(&istr,20));
  if ( buf[1] || ! schd->ug2g(buf, &(g_ot)) ) {
    wlog3 ("Incorrect code for Over-Time \"", buf, "\" on line: \n");
    wlog2 (rec, "\n  ABORT.\n");
    wfexit(1);
  }

  strcpy (buf, get_s(&istr,20));
  if ( buf[1] || ! schd->ug2g(buf, &(g_ut)) ) {
    wlog3 ("Incorrect code for Under-Time \"", buf, "\" on line: \n");
    wlog2 (rec, "\n  ABORT.\n");
    wfexit(1);
  }

  strcpy (buf, get_s(&istr,20));
  seniority  = atoi(buf);
  if (seniority < 1 || seniority > 99) {
    wlog3 ("Incorrect Seniority  \"", buf, "\" on line: \n");
    wlog2 (rec, "\n  (must be between 1 and 99).  ABORT.\n");
    wfexit(1);
  }

  int low = seniority;
  if (low > 10) low = 10;
  int med = seniority - low;
  if (med > 30) med = 30;
  int high = seniority - (med + low);

  i_seniority = 32 + low + (med / 2) + (high / 4);

  strcpy (buf, get_s(&istr,20));
  if (buf[0] == 'Y' || buf[0] == 'y') {
    strcpy (vary_shfts, "Y");
  } else if (buf[0] == 'N' || buf[0] == 'n') {
    strcpy (vary_shfts, "N");
  } else {
    wlog3 ("Incorrect Vary-Shfts  \"", buf, "\" on line: \n");
    wlog2 (rec, "\n  (must be Y or N).  ABORT.\n");
    wfexit(1);
  }

}

//---------------------------------------------------------
//  Setup before we attempt a completion.

//  We are either just about to start for the first time or we have
//  just finished backing-out from a previous attempt.  cur_day
//  points to the first day which is not done.

void  employee::try_setup (void)
{
  if (badge != Badge_Id) abort ("Emp-ts1");

  short int iday;

  days_on = days_off = 0;
  for (iday=0; iday<cur_day->iday; iday++) {     // Previous week.
    empl_day *d = day + iday;
    if ( ! d->done ) abort ("Emp-ts2");
    if (d->cur_fill->day_off) {
      days_on = 0;
      days_off++;
    } else {
      days_off = 0;
      days_on++;
    }
  }

  days_can_work = days_can_off = 0;
  for (iday=cur_day->iday; iday<Num_Days; iday++) {   // Current week.
    empl_day *d = day + iday;
    if (d->done) abort ("Emp-ts3");
    d->try_setup();
    if (d->can_work) days_can_work++;
    if (d->can_off)  days_can_off++;
  }

  if (days_can_work < 1) {
    wlog3 ("Missing or bad details for ", name, "\n");
    wlog  ("\nAbort\n");
    wfexit (1);
  }

  shft_last = null;
  empl_day *ed = day + Week_Start;
  while (ed->iday > 0) {     // Find last day worked.
    ed--;
    if (ed->cur_fill) {
      if (ed->cur_fill->shft) {
        shft_last = ed->cur_fill->shft;
        break;
      }
    } else {
      abort ("Emp-ts4");
    }
  }

  day_setup();     // Prepare to do a day.

}

//---------------------------------------------------------
//  Setup before starting to do a new day.

void  employee::day_setup (void)
{
  if (badge != Badge_Id) abort ("Emp-ds1");

  eval_work_today();   // Evaluate hope and goodness of working today.

  cur_day->day_setup();

  det_imp();            // Determine importance, move in list.

}

//---------------------------------------------------------
// Apply the passed fill to this emp.

void  employee::do_fill (fill *f)
{
  if (badge != Badge_Id) abort ("Emp-df1");

  if (f->emp->iemp != iemp) abort ("Emp-df2");
  if (done) abort ("Emp-df3");
  if (f->iday != cur_day->iday) abort ("Emp-df4");

  empl_day *ed = day + f->iday;     // The day of the fill.

  ed->do_fill(f);

  g += f->g;

  if (!f->day_off) min_ttp += f->shft->minutes;

  if (ed->can_work) days_can_work--;
  if (ed->can_off)  days_can_off--;
  if (f->day_off) {
    days_off++;
    days_on = 0;
    days_off_need--;
  } else {
    days_on++;
    days_off = 0;
    days_on_need--;
  }

  if (ed->iday == Num_Days-1) {
    done = True;
    det_imp();
  } else if (ed->done) {
    cur_day++;
    day_setup();
  }

  if (f->shft) shft_last = f->shft;

  i_left_bhnd -= 256;        // Being "left behind" less.

}

//---------------------------------------------------------
// Un-apply the passed fill to this emp.

void  employee::undo_fill (fill *f)
{
  if (badge != Badge_Id) abort ("Emp-uf1");

  if (!f->emp || f->emp != this) {
    wlog ("Internal error: Bad emp in emp::undo_fill().  Abort.\n");
    wfexit(1);
  }

  empl_day *ed = day + f->iday;

  ed->undo_fill (f);

  g -= f->g;

  if (!f->day_off) min_ttp -= f->shft->minutes;

  done = False;

  cur_day = ed;      // This is the first day that isn't done.

  i_left_bhnd += 256;        // Being "left behind" more.

  if (ed->can_work) days_can_work++;
  if (ed->can_off)  days_can_off++;
  if (f->day_off) {
    days_off_need++;
  } else {
    days_on_need++;
  }

}

//---------------------------------------------------------
//  Determine the importance of this emp, set it in the
//  data, and, if the emp is part of a schedule, move the
//  emp in the list so as to keep the list in decreasing
//  order of importance.


void  employee::det_imp (void)
{
  if (badge != Badge_Id) abort ("Emp-di1");

  if (done) {

    imp = -999999;

  } else if (one_choice()) {

    imp = 99999 + i_seniority;

  } else {

    det_i_poor_shfts();
    imp = i_been_prob + i_poor_shfts;
    if (h_work_today > Good || h_work_today <= Bad) {
      imp += (Good * 4);
    }
    imp += i_left_bhnd;      // Favors scheduling by days.
    imp += i_strategic;      // Cyclically mix things up a little.
    imp += cur_day->i_been_prob;
    imp += cur_day->i_imp_fills;
    imp += i_seniority;

  }

  #define  prev_OLElm  prev_emp
  #define  next_OLElm  next_emp
  #define  OLKey       imp
  #define  OLHead      (schd->emp)
  #define  OLElmTyp    employee

  if (schd) {                      // Ensure that most important is first.
    #include "olsetkey.inc"
  }

  #undef  prev_OLElm
  #undef  next_OLElm
  #undef  OLKey
  #undef  OLHead
  #undef  OLElmTyp

}

//---------------------------------------------------------
// Determine goodness and hope of working and not working on cur_day.
// Goodness is used to affect the goodness of the schedule.
// Hope is used to decide if we should do it.

// In this first version we will detect various problems only as
// they begin to happen; should anticipate running out of days to
// take off, for instance.

void  employee::eval_work_today (void)
{
  if (badge != Badge_Id) abort ("Emp-ewt1");

  if (done) abort ("Emp-ewt2");

  if (must_work()) {
         // Emp must work if all the shft-days are to be satisfied.
    g_work_today = Ok;
    g_off_today  = Imposible;
    h_work_today = Manditory;
    h_off_today  = Imposible;
    return;
  }

  goodness g_o;
  if (g_ot >= Expected) {
    g_o = Poor;
  } else if (g_ot >= Poor) {
    g_o = (Poor + Bad) / 2;
  } else if (g_ot < Poor) {
    g_o = Bad;
  }

  goodness g_u;
  if (g_ut >= Expected) {
    g_u = Poor;
  } else if (g_ut >= Poor) {
    g_u = (Poor + Bad) / 2;
  } else if (g_ut < Poor) {
    g_u = Bad;
  }
                                                 // Emp should not work...

  if (days_on_need <= 0) {
         // Emp has already worked enough this week.

    if (days_on_need == 0) {
      g_work_today = g_o;
      g_off_today  = Ok;
      h_work_today = g_o;
      h_off_today  = g_o * -1;
    } else {
      g_work_today = g_o * 4;
      g_off_today  = Ok;
      h_work_today = g_o * 4;
      h_off_today  = g_o * -4;
    }

  } else if (days_off_need >= days_can_off) {
         // If emp works today, there will not be enough days he can take
         // off later to prevent overtime.

    if (days_off_need == days_can_off) {
      g_work_today = Ok;      // Will be caught when it becomes a problem.
      g_off_today  = Ok;
      h_work_today = g_o;
      h_off_today  = g_o * -1;
    } else {
      g_work_today = Ok;
      g_off_today  = Ok;
      h_work_today = g_o * 4;
      h_off_today  = g_o * -4;
    }

  } else if (days_on >= days_per_week) {
         // If emp works today, he will work too many days in a row.

    if (days_on >= Week_Len-1) {
      g_work_today = Bad;
      g_off_today  = Ok;
      h_work_today = Bad;
      h_off_today  = Bad * -1;
    } else if (days_on == days_per_week) {
      g_work_today = g_o;
      g_off_today  = Ok;
      h_work_today = g_o;
      h_off_today  = g_o * -1;
    } else {
      g_work_today = g_o * 4;
      g_off_today  = Ok;
      h_work_today = g_o * 4;
      h_off_today  = g_o * -4;
    }
                                                 // Emp should work...

  } else if (days_off_need <= 0) {
         // Emp has already had enough days off this week.

    if (days_off_need == 0) {
      g_work_today = Ok;
      g_off_today  = g_u;
      h_work_today = g_u * -1;
      h_off_today  = g_u;
    } else {
      g_work_today = Ok;
      g_off_today  = g_u *  4;
      h_work_today = g_u * -4;
      h_off_today  = g_u *  4;
    }

  } else if (days_on_need >= days_can_work) {
         // If emp takes today off, there will not be enough days
         // he can work later to prevent undertime.

    if (days_on_need == days_can_work) {
      g_work_today = Ok;
      g_off_today  = Ok;   // Will be caught when it becomes a problem.
      h_work_today = g_u * -1;
      h_off_today  = g_u;
    } else {
      g_work_today = Ok;
      g_off_today  = Ok;
      h_work_today = g_u * -4;
      h_off_today  = g_u * 4;
    }

  } else if (days_off >= off_per_week) {
         // If emp takes today off, he will take too many days off in a row.

    if (days_off == off_per_week) {
      g_work_today = Ok;
      g_off_today  = g_u;
      h_work_today = g_u * -1;
      h_off_today  = g_u;
    } else {
      g_work_today = Ok;
      g_off_today  = g_u *  2;
      h_work_today = g_u * -2;
      h_off_today  = g_u *  2;
    }

  } else if (days_off > 0) {    // Emp had yesterday off.

    g_work_today = Poorish;
    g_off_today  = Ok;
    h_work_today = Poorish;
    h_off_today  = Good;

  } else {

   g_work_today = Ok;
   g_off_today  = Ok;
   h_work_today = (Ok + Good) / 2;
   h_off_today  = Poorish;

  }

}

//----------------------------------------------------------
//  A try has finished.

void  employee::note_try (void)
{
  if (badge != Badge_Id) abort ("Emp-nt1");

  if (g > g_prev) {
    g_try = Good;
  } else if (g <= Bad) {
    g_try = Bad;
  } else if (g < Ok) {
    g_try = Poor;
  } else {
    g_try = Ok;
  }
  if (done       && g_try >= Ok) g_try += 1;
  if (schd->done && g_try >= Ok) g_try += 1;

  if (schd->fail && schd->fail->emp == this) num_fail++;

  if (schd->fail && schd->fail->emp == this) {
    if (i_been_prob == 32) {
      i_been_prob = 0;
    } else {
      i_been_prob = 32;
    }
  } else if (g < Poor || g_try < Poor) {
    if (i_been_prob > 64) {
      i_been_prob = 0;
    } else {
      i_been_prob += 16;
    }
  }

  if (schd->worst_fill && schd->worst_fill->emp->iemp == iemp) {
    i_been_prob += 4;
  }

  short int iday;
  for (iday=Week_Start; iday<Num_Days; iday++) {
    empl_day *ed = day + iday;
    ed->note_try();
  }

  if (++i_strategic > 2) i_strategic = 0;    // Roll it.

  g_prev = g;

}

//----------------------------------------------------------
//  Remember that the current situation as part of the
//  best-so-far schedule.

void  employee::save_bsf (void)
{
  if (badge != Badge_Id) abort ("Emp-sb1");

  if (!done) abort ("Emp-sb2");

  short int iday;
  for (iday=Week_Start; iday<Num_Days; iday++) {  // For each day...
    empl_day *ed = day + iday;
    ed->save_bsf();
  }

  g_bsf = g;

}

//----------------------------------------------------------
//  I have little hope for improvement; I must change the way I look at
//  things so as to try different things.

void  employee::broaden_horizons (void)
{
  if (badge != Badge_Id) abort ("Emp-bh1");

  int iday;
  for (iday=Week_Start; iday<Num_Days; iday++) {
    empl_day *ed = day + iday;
    ed->broaden_horizons();
  }

  if (g_bsf >= Ok) {
    i_been_prob /= 2;
  } else if (g_bsf <= Bad) {
    i_been_prob *= 2;
  } else {
    i_been_prob = (i_been_prob * 3) / 2;   // Lighten up a little.
  }

  short int ibase = iemp + schd->itry + schd->num_fail;
  i_strategic = ibase - ((ibase / 3) * 3);  // High value is 4.
//  i_strategic = rand() % 3;       // High value is 2.

}

//---------------------------------------------------------
//  Check the data-structure integrity.

void  employee::check_integrity (void)
{
  if (badge != Badge_Id) abort ("Emp-Badge");

  empl_day *d = day;
  int num_d = 0;
  int can_work = False;
  while (d) {
    if (++num_d > Num_Days) abort ("Emp-16");
    can_work = can_work || d->num_pos_fill > 1;
    d->check_integrity();
    d = d->next_day;
    if (d) {
      if (d->prev_day->next_day->iday != d->iday) abort ("Emp-17");
    }
  }
  if (num_d != Num_Days) abort ("Emp-18");

  if (days_per_week > 0 && ! can_work) {
    wlog3 ("\nThere are no Details for employee: ", name, "   Abort.\n");
    wfexit(1);
  }

}

//------------------------------------------------------------------------
//  Write a line to the log identifying this emp and indicating
//  why it is important.

void  employee::log_imp (void)
{
  if (badge != Badge_Id) abort ("Emp-Badge");

  char str1[101], str2[101];

  sprintf (str1, "E%2d: %2dD %7ldI| %2ldbp %2lddbp %2ldsen %2ldlb",
           int(iemp), int(cur_day->iday), imp, i_been_prob,
           cur_day->i_been_prob, i_seniority, i_left_bhnd);

  sprintf (str2, " %4ldps %4ldstgc %4ldwrk %4ldoff %4ldif",
           i_poor_shfts, i_strategic, h_work_today, h_off_today,
           cur_day->i_imp_fills);

  wlog3 (str1, str2, "\n");

}

//------------------------------------------------------------------------
//  Return True if the employee has only one pos_fill (or none)  which
//  could be worked on the current day.

int  employee::one_choice (void)
{
  if (badge != Badge_Id) abort ("Emp-Badge");

  short int iday = cur_day->iday;

  short int num = 0;

  fill *f;
  short int i;
  for (i=0; i<cur_day->num_pos_fill; i++) {
    f = cur_day->pos_fill_array + i;

    if (f->g_static <= schd->g_cutoff) continue;
    if (f->shft) {
      shft_day *sd = f->shft->day + iday;
      if (sd->done) continue;
      if (g_work_today <= schd->g_cutoff) continue;
    } else {         // This fill is for a day-off.
      if (g_off_today  <= schd->g_cutoff) continue;
    }

    num++;

  }

  return (num <= 1);

}

//------------------------------------------------------------------------
//  Return True if:
//    the number of employees who can still work on this emp's cur_day
//    is <= the number of employees still required by shifts this day.


int  employee::must_work (void)
{
  if (badge != Badge_Id) abort ("Emp-Badge");

  if (done || cur_day->done) return False;

  employee  *e;
  empl_day  *ed;
  shift     *s;
  shft_day  *sd;
  fill      *f;

  short int iday = cur_day->iday;

  short int this_can = False;

  f = cur_day->pos_fill;
  for (;f;f=f->next_fill) {    // Each fill e can do on this.cur_day...
    if (f->g_static <= schd->g_cutoff) continue;
    if (f->shft) {
      this_can = True;     // This emp can work on cur_day.
      break;
    }
  }

  if (!this_can) return False;

  short int num_can = 0;

  e = schd->emp;
  for (;e;e=e->next_emp) {
    ed = e->day + iday;
    if (ed->done) continue;
    f = ed->pos_fill;
    for (;f;f=f->next_fill) {    // Each fill e can do on this.cur_day...
      if (f->g_static <= schd->g_cutoff) continue;
      if (f->shft) {
        num_can++;
        break;
      }
    }
  }

  short int num_req = 0;

  s = schd->shft;
  for (;s;s=s->next_shft) {
    sd = s->day + iday;
    num_req += (sd->emps_req - sd->emps_have);
  }

  return (num_req > 0 && num_can <= num_req);

}

//------------------------------------------------------------------------
//  Determine i_poor_shfts - the goodness factor intended to increase
//  the importance of employees that already have fills of poor shifts
//  so that they get a better chance of getting better shifts later in
//  the schedule.

void  employee::det_i_poor_shfts (void)
{
  if (badge != Badge_Id) abort ("Emp-Badge");

  fill *f;
  goodness gs;

  i_poor_shfts = 0;

  empl_day  *ed = day + Week_Start;
  while (ed) {                         // For all days...
    f = ed->cur_fill;                  
    if (!f) break;                       // which are done...
    gs = f->g_static - f->emp_day->g_day;
    if (gs <= Bad) {
      i_poor_shfts += 8;
    } else if (f->g_shft <= Poor) {
      i_poor_shfts += 4;
    }
    ed = ed->next_day;
  }

}

//---------------------------------------------------------
