//  Brian Marshall      bmarshal@agt.net
//......................................

//  empl_day member functions.

#include <stdlib.h>
#include "workforc.h"

//----------------------------------------------------------
//  Setup this employee day so we can begin solving the problem.

void  empl_day::problem_setup (void)
{
  if (badge != Badge_Id) abort ("Eday-ps1");

  if (iday < Week_Start) {
    pos_fill_array = new fill[1];
  } else {                        // Every shift plus the day off.
    pos_fill_array = new fill[emp->schd->num_shft+1];
  }
  if ( ! pos_fill_array ) {
    wlog ("Error getting memory for pos_fill.\n");
    wfexit(1);
  }
  pos_fill = pos_fill_array;

  num_pos_fill   = 0;

  g_day = Ok;

  new_pos_fill (-1, Ok, Ok);  // Day-off is first or default pos_fill.

  if (iday < Week_Start) {
    cur_fill = pos_fill;
    done = True;
  } else {
    cur_fill = null;
    done = False;
  }

  i_been_prob = Trivial;

  num_fail = 0;

  bsf_fill = null;
  g_bsf = Imposible;

  if (bsf_fill) abort ("Eday-3");
  if (num_fail) abort ("Eday-4");

}

//------------------------------------------------------------------
// Setup before we can attempt a completion.

void  empl_day::try_setup (void)
{
  if (badge != Badge_Id) abort ("Eday-ts1");

  can_work = can_off = False;
  short int ifill;
  for (ifill=0; ifill<num_pos_fill; ifill++) {
    fill *f = pos_fill_array + ifill;

    if (f->day_off) {
      if (f->g_static > emp->schd->g_cutoff) can_off = True;
    } else {
      if (f->g_static > emp->schd->g_cutoff) {
        can_work = True;
        shft_day *sd = f->shft->day + iday;
        sd->note_could();  // empl-day will be able to do this shft-day.
      }
    }

  }

}

//------------------------------------------------------------------
// Setup before we can "do" this day (this day has just become cur_day).

void  empl_day::day_setup (void)
{
  if (badge != Badge_Id) abort ("Eday-ds1");

  short int ifill;
  i_imp_fills = Trivial;
  for (ifill=0; ifill<num_pos_fill; ifill++) {   // For each posible fill...
    fill *f = pos_fill_array + ifill;
    f->day_setup();     // Get it ready now that it is on cur_day.
    if (f->hope > Good) i_imp_fills += 2;
    if (! f->day_off ) {
      if (f->g_static > emp->schd->g_cutoff) {
        shft_day *sd = f->shft->day + f->iday;
        sd->note_can();      // Now can work this shft_day.
      }
    }
  }    // End of loop through posible fills.

}

//------------------------------------------------------------------
//  Apply this fill to this empl-day.

void  empl_day::do_fill (fill *f)
{
  if (badge != Badge_Id) abort ("Eday-df1");

  cur_fill = f;

  f->do_fill();

  done = True;

  fill *pf = pos_fill;
  while (pf) {                    // For each pos_fill...
    if ( ! pf->day_off ) {
      shft_day *sd = pf->shft->day + pf->iday;
      sd->note_busy();      // No longer free to do anything else today.
    }
    pf = pf->next_fill;
  }

}

//------------------------------------------------------------------
//  Un-apply this fill to this empl-day.

void  empl_day::undo_fill (fill *f)
{
  if (badge != Badge_Id) abort ("Eday-uf1");

  cur_fill = null;

  done = False;

  f->undo_fill();

}

//----------------------------------------------------------
//  Remember that the current fill is part of the best-so-far schedule.

void  empl_day::save_bsf (void)
{
  if (badge != Badge_Id) abort ("Eday-sb1");

  if (!cur_fill || cur_fill->badge != Badge_Id) abort ("Eday-sb2");

  bsf_fill = cur_fill;
  g_bsf = bsf_fill->g;

}

//----------------------------------------------------------
//  A try has finished.

void  empl_day::note_try (void)
{
  if (badge != Badge_Id) abort ("Eday-nt1");

  if (cur_fill) {
    cur_fill->note_try();
  }

  schedule *z = emp->schd;
  if (z->fail) {
    if (z->fail->emp == emp && z->fail->iday == iday) {
      num_fail++;
      i_been_prob += 16;
    }
  } else {
    if (emp->schd->g > Imposible && i_been_prob > 16) {
      i_been_prob -= 4;
    }
  }

}

//----------------------------------------------------------
//  I have little hope for improvement; I must change the way I look at
//  things so as to try different things.

void  empl_day::broaden_horizons (void)
{
  if (badge != Badge_Id) abort ("Eday-bh1");

  int ifill;
  for (ifill=0; ifill<num_pos_fill; ifill++) {
    fill *f = pos_fill_array + ifill;
    f->broaden_horizons();
  }

  i_been_prob /= 2;

}

//------------------------------------------------------------------
//  Add a new posible-fill to this empl_day.
//  gs  is the goodness specified for the shift in general, while
//  gsd is the goodness specified for the shift on this day.

//  If this empl_day is for last-week and, a pos_fill already exists,
//  over-write it (it is a "default").

void  empl_day::new_pos_fill (short int ishft, goodness gs, goodness gsd)
{
  if (badge != Badge_Id) abort ("Eday-npf1");

  if (iday >= Week_Start || num_pos_fill == 0) num_pos_fill++;

  fill *f = pos_fill_array + num_pos_fill - 1;    // The fill.

  f->badge = Badge_Id;

  f->ifill = emp->iemp * (emp->schd->num_shft+1) + num_pos_fill - 1;

  f->next_fill   = null;
  if (num_pos_fill == 1) {
    f->prev_fill = null;
  } else {
    f->prev_fill = f - 1;
    (f->prev_fill)->next_fill = f;
  }
  f->emp_day = this;

  f->problem_setup (ishft, gs, gsd);

}

//---------------------------------------------------------
//  Check the data-structure integrity.

void  empl_day::check_integrity (void)
{
  if (badge != Badge_Id) abort ("Eday-Badge");

  if (num_pos_fill < 1 ||
      num_pos_fill > emp->schd->num_emp+1) abort ("Eday-1");
  if (iday < emp->cur_day->iday) {
    if ( ! cur_fill ) abort ("Eday-5");
    if (cur_fill->iday != iday ||
        cur_fill->emp->iemp != emp->iemp) abort ("Eday-6");
    if ( ! done ) abort ("Eday-11");
  } else {
    if (cur_fill) abort ("Eday-7");
    if (done) abort ("Eday-2");
  }

  fill *f = pos_fill;
  int num_f = 0;
  while (f) {
    if (++num_f > num_pos_fill) abort ("Eday-8");
    if (f->emp->iemp != emp->iemp) abort ("Eday-11");
    if (f->iday != iday) abort ("Eday-12");
    f->check_integrity();
    f = f->next_fill;
    if (f) {
      if (f->prev_fill->next_fill->iday != f->iday) abort ("Eday-9");
    }
  }
  if (num_f != num_pos_fill) abort ("Eday-10");

}

//---------------------------------------------------------
//  Set the goodness associated with the employee working this day.

void  empl_day::set_g_day (goodness a_g_day)
{
  if (badge != Badge_Id) abort ("Eday-Badge");

  g_day = a_g_day;

}

//------------------------------------------------------------------
