//  Brian Marshall      bmarshal@agt.net
//......................................

//  Fill functions.

#include <stdlib.h>
#include "workforc.h"

//----------------------------------------------------------
//  Flags controlling how much detail gets written to the log.
//  We use ((emp->schd->num_shft)) and (!(emp->schd->num_shft))
//  so that the compiler never complains that code is not reachable.

                          // Log application of rotations to this fill?
#define     Log_Rotations   ((emp->schd->num_shft))

//---------------------------------------------------------
//  Set up this fill as something that we now know it's emp can do.
//  gs  is the goodness specified for the shift in general, while
//  gsd is the goodness specified for the shift on this day.
//  Note that all fills are initially given the same hope so that, no
//  matter what order they are created, they will not be out of order.

void  fill::problem_setup (short int ishft, goodness gs, goodness gsd)
{
  if (badge != Badge_Id) abort ("Fill-ps1");

  if (ishft < 0) {
    day_off = True;
    shft    = null;
  } else {
    day_off = False;
    shft    = emp_day->emp->schd->shft + ishft;
  }
  emp         = emp_day->emp;
  iday        = emp_day->iday;
  g_shft      = gs;

  if (gsd == Neutral) {
    g_static = Ok + gs  + emp_day->g_day;
  } else {
    g_static = Ok + gsd + emp_day->g_day;
  }

  g_rotation = Neutral;

  g           = g_static;
  g_shft_chg  = Ok;
  hope        = 1;

  if (g_static >= Expected) {
    h_grss_grnr = 8;
  } else {
    h_grss_grnr = 0;
  }
  h_grss_grnr_x = 1;  // Decrement (rather than halve) next time.

  h_wrks_well = h_been_prob = 0;
  has_worked = has_worked_emp = False;

}

//---------------------------------------------------------
// The emp of this fill has gotten up to the day of this fill and we now
// wish to get this posible fill ready to be considered.

void  fill::day_setup (void)
{
  if (badge != Badge_Id) abort ("Fill-ds1");

  det_g_shft_chg();   // Determine goodness considering yesterday.

  if (day_off) {
    g_work_today = emp->g_off_today;
  } else {
    g_work_today = emp->g_work_today;
  }

  evaluate_mpw ();         // Determine minutes-per-week goodness and hope.

  det_g();        // Determine goodness of this fill.

  det_hope();     // Determine hope of this fill and adjust position in list.

}

//---------------------------------------------------------
//  Apply the effect of doing this fill to this fill.

void fill::do_fill (void)
{
  if (badge != Badge_Id) abort ("Fill-df1");
}

//---------------------------------------------------------
//  Un-apply the effect of doing this fill to fill.

void fill::undo_fill (void)
{
  if (badge != Badge_Id) abort ("Fill-uf1");
}

//---------------------------------------------------------
//  A try including this fill has finished.

void  fill::note_try (void)
{
  if (badge != Badge_Id) abort ("Fill-nt1");

                 
  if (emp->schd->tries_since_bsf > 3) {
    if (h_grss_grnr_x == 0) {
      h_grss_grnr /= 2;  // "Grass is greener" decreases as fill is tried.
      h_grss_grnr_x++;
    } else {
      h_grss_grnr_x--;   // every second try.
    }
  }

  employee *e = emp_day->emp;
  schedule *z = e->schd;

  if (z->done) {
    if (z->g >= Ok) has_worked = True;
    if (e->g >= Ok) has_worked_emp = True;
  }
  if (h_wrks_well >= 8) h_wrks_well /= 2;
  if (e->done && e->g >= Ok) {
    if (z->g >= Ok) {
      h_wrks_well += 2;
    } else {
      h_wrks_well += 1;
    }
  }
  if (e->g_try > Ok) {
    h_wrks_well += (e->g_try / 2);
  }
  if (z->g_try > Ok) h_wrks_well += 1;
  if (h_wrks_well > 2) h_wrks_well--;

  if (g <= Bad) {                              // Really bad fill.
    if (h_been_prob > -12) h_been_prob -= 4;
  } else if (g <= Poor) {                      // Bad fill.
    if (h_been_prob > -12) h_been_prob -= 2;
  }
  if (e->g <= Bad) {                            // Really bad for emp.
    if (h_been_prob > -12) h_been_prob -= 4;
  } else if (e->g <= Poor) {                    // Bad for emp.
    if (h_been_prob > -12) h_been_prob -= 2;
  }
  if (z->g_try <= Bad || e->g_try <= Bad) {    // Really bad try.
    if (h_been_prob > -12) h_been_prob -= 2;
  } else if (z->g_try < Ok || e->g_try < Ok) { // Bad try.
    if (h_been_prob > -12) h_been_prob -= 1;
  }
  if (h_been_prob < -2) h_been_prob++;      // Slowly forget.

}

//---------------------------------------------------------
//  Determine the goodness of this fill.

//  Note that seniority is only about bad stuff.

void  fill::det_g (void)
{
  if (badge != Badge_Id) abort ("Fill-dg1");

  goodness g1;

  g = g_static;

  g1 = g_shft_chg;
  if (g1 < 0) g1 = ((g1 * emp->i_seniority) / Imp_Divsr);
  g += g1;

  g1 = g_work_today;
  if (g1 < 0) g1 = ((g1 * emp->i_seniority) / Imp_Divsr);
  g += g1;

  g1 = g_mpw;   // Apply emp's minutes-per-week aspects.
  if (g1 < 0) g1 = ((g1 * emp->i_seniority) / Imp_Divsr);
  g += g1;

}

//---------------------------------------------------------
//  Determine the hope of this fill, set it in the
//  data, and, if the fill is part of an empl_day, move the
//  fill in the list so as to keep the list in decreasing
//  order of hope.

//  Note that g has already been affected by emp->g_work_today
//  if it is bad; I have to apply it to hope if it is good.

void  fill::det_hope (void)
{
  if (badge != Badge_Id) abort ("Fill-dh1");

  hope = (g * 2) + h_wrks_well + h_been_prob + h_mpw + h_grss_grnr;

  if (day_off) {
    hope += emp->h_off_today;
  } else {
    hope += emp->h_work_today;
  }

  if (emp_day->bsf_fill && ifill == emp_day->bsf_fill->ifill) {
    hope += 2;
    if (emp->schd->g_bsf >= Ok) hope += 2;
  } else if (has_worked) {
    hope += 2;
  }
  if (has_worked_emp) hope += 1;

  if (shft) {                      
//    hope += (shft->i_been_prob / 4)??;
    shft_day *sd = shft->day + iday;
    hope += sd->i_been_prob / 4;     // Has shft_day been hard to fill?
  }

  #define  prev_OLElm  prev_fill
  #define  next_OLElm  next_fill
  #define  OLKey       hope
  #define  OLHead      (emp_day->pos_fill)
  #define  OLElmTyp    fill

  if (emp_day) {                     // Ensure that most hopefull is first.
    #include "olsetkey.inc"
  }

  #undef  prev_OLElm
  #undef  next_OLElm
  #undef  OLKey
  #undef  OLHead
  #undef  OLElmTyp

}

//----------------------------------------------------------
//  I have little hope for improvement; I must change the way I
//  look at things so as to try different things.

void  fill::broaden_horizons (void)
{
  if (badge != Badge_Id) abort ("Fill-bh1");

  if (h_wrks_well == 0 && h_been_prob == 0) {  // Never tried.
    h_grss_grnr = 16;
  } else if (this == emp_day->bsf_fill) {
    h_grss_grnr = 4;
  } else {
    h_grss_grnr = 8;
  }
  h_grss_grnr_x = 1;

  h_wrks_well /= 2;
  h_been_prob /= 2;

}

//---------------------------------------------------------
//  Don't consider doing this fill.

void  fill::reject (void)
{
  if (badge != Badge_Id) abort ("Fill-r1");
  g = g_static = Imposible;
}

//------------------------------------------------------------------
//  Check the data-structure integrity.

void  fill::check_integrity (void)
{
  if (badge != Badge_Id) abort ("Fill-Badge");

  if (emp_day->iday != iday ||
      emp_day->emp->iemp != emp->iemp) abort ("Fill-4");
  if (shft && day_off) abort ("Fill-1");
  if ( ! shft && ! day_off) abort ("Fill-2");
}

//------------------------------------------------------------------------
//  Write a line to the log identifying this fill and indicating
//  why it is hopefull.

void  fill::log_hope (void)
{
  if (badge != Badge_Id) abort ("Fill-Badge");

  char str1[101], str2[101];

  if (shft) {
    sprintf (str1, "....%2dE %2dS %2dD %7ldH %7ldG|",
             int(emp->iemp), int(shft->ishft), int(iday), hope, g);
  } else {
    sprintf (str1, "....%2dE  -- %2dD %7ldH %7ldG|",
             int(emp->iemp), int(iday), hope, g);
  }
  sprintf (str2, " %4ldgg %4ldww %4ldbp %4ldm\n",
           h_grss_grnr, h_wrks_well, h_been_prob, h_mpw);

  wlog2 (str1, str2);

}

//------------------------------------------------------------------------
//  Determine g_shft_chg of this fill.
//
//  g_shft_chg is the goodness of working this shift considering
//  what was done yesterday (and, in relation to the employee's
//  desire to work varying shifts, the rest of the week up to this
//  day).

void  fill::det_g_shft_chg (void)
{
  if (badge != Badge_Id) abort ("Fill-Badge");

  fill *yf = (emp_day-1)->cur_fill;          // Yesterday's fill.

  if (!shft) {                    // This is a day off.

    if (!yf->shft) {                  // Also had yesterday off.
      g_shft_chg = Goodish;
    } else {                          // Worked yesterday.
      g_shft_chg = Ok;     
    }
                    
  } else {                        // This is not a day off.

    if (!yf->shft) {                  // Had yesterday off.

      g_shft_chg = Expected;
      if (emp->shft_last) {    // Know what previous shift was.
        if (shft->ishft == emp->shft_last->ishft) {  // Same one.
          g_shft_chg -= 2;   // Rotate: Favor a different shift.
        }
      }


    } else {                          // Also worked yesterday...

      if (shft->ishft == yf->shft->ishft) {   // Same shift yesterday.

        if (emp->vary_shfts[0] == 'Y') {
          if (yf->g_shft >= Expected) {
            g_shft_chg = Poorish;            
          } else {
            g_shft_chg = Poor;
          }
        } else {   // Emp does not want varying shifts.
          if (yf->g_shft >= Expected) {
            g_shft_chg = Goodish;
          } else {
            g_shft_chg = Ok;
          }
        }

      } else {                                // Different shift yesterday. 

        int brk = (2400 + shft->start_time) - yf->shft->end_time;
        if (brk > 2400) brk -= 2400;
        if (brk < 900) {               // Break less than 9 hours.
          g_shft_chg = Very_Bad;
        } else if (brk < 1100) {       // Break less than 11 hours.
          g_shft_chg = Bad;
        } else if (brk < 1300) {       // Break less than 13 hours.
          g_shft_chg = Poor;
        } else {                       // Not excessively short break.
          if (emp->vary_shfts[0] == 'Y') {
            g_shft_chg = Goodish;
          } else {
            if (g_shft > yf->g_shft) {   // Better than yesterday.
              g_shft_chg = Poorish;
            } else {                     // Just a different shift.
              g_shft_chg = Poor;
            }
          }
        }

      }     // End of block for different shift yesterday.

    }     // End of block for worked yesterday and today.

    if (emp->vary_shfts[0] == 'Y') {  // Emp wants varying shifts.
      fill *past_fill;
      short int jday;
      for (jday=Week_Start; jday<iday-1; jday++) {
        past_fill = emp->day[jday].cur_fill;
        if (past_fill->shft && past_fill->shft->ishft == shft->ishft) {
          g_shft_chg += Poorish;  // Already worked this shift this week.
        }
      }
    }

  }     // End of block for working today.

}

//------------------------------------------------------------------------
//  Determine g_mpw and h_mpw of this fill.
//
//  g_mpw and h_mpw are the goodness and hope of this fill considering
//  minutes-per-week aspects.

void  fill::evaluate_mpw (void)
{
  if (badge != Badge_Id) abort ("Fill-Badge");

  goodness  g_o_base, g_o_div;    // Used to determine goodnesses.
  if (emp->g_ot >= Expected) {
    g_o_base = -1;
    g_o_div = 64;
  } else if (emp->g_ot == Poor) {
    g_o_base = -4;
    g_o_div = 32;
  } else if (emp->g_ot >= Bad) {
    g_o_base = -16;
    g_o_div = 8;
  } else {
    g_o_base = -32;
    g_o_div = 8;
  }

  goodness  g_u_base, g_u_div;    // Used to determine goodnesses.
  if (emp->g_ut >= Expected) {
    g_u_base = -1;
    g_u_div = 64;
  } else if (emp->g_ut == Poor) {
    g_u_base = -4;
    g_u_div = 32;
  } else if (emp->g_ut >= Bad) {
    g_u_base = -16;
    g_u_div = 8;
  } else {
    g_u_base = -32;
    g_u_div = 8;
  }

  short int minutes = emp->min_ttp;           // Minutes to this point
  if (! day_off) minutes += shft->minutes;    //   including this fill.

  short int last_day = iday == Num_Days-1 ||
                       (emp->days_on_need == 1 && ! day_off);

                  // Check for actual minutes-per-week problem.
  g_mpw = Ok;
  if (emp->min_per_week > 0) {               // Care about minutes/week?
    short int  diff = minutes - emp->min_per_week;
    if (diff > 0 && ! day_off) {
      g_mpw = g_o_base - diff / g_o_div;           // Over-time.
    } else if (last_day && diff < 0) {
      g_mpw = g_u_base - (diff * -1) / g_u_div;    // Under-time.
    }
  }
                  // Check for potential minutes-per-week problem.
  h_mpw = Ok;
  if (emp->min_per_week > 0) {               // Care about minutes/week?
    if (!day_off && emp->days_on_need > 1) {
      short int  projected_minutes = minutes +
                                     (emp->days_on_need-1) * shft->minutes;
      short int  diff = projected_minutes - emp->min_per_week;
      if (diff > 0) {
        h_mpw = g_o_base - diff / g_o_div;           // Over-time.
      } else if (diff < 0) {
        h_mpw = g_u_base - (diff * -1) / g_u_div;    // Under-time.
      } else {
        h_mpw = Ok;
      }
      h_mpw = (h_mpw * emp->i_seniority) / Imp_Divsr;
      h_mpw = (h_mpw * 2) / emp->days_on_need;
      h_mpw = (h_mpw * 2) / 3;
    }
  }

}

//------------------------------------------------------------------------
//  Apply a rotation to this fill.
//
//  The effect of emp_day->g_day is removed and either Ok or
//  g_if_fail is added to g_static.
//
//  a_shft_str is a shift-name, On or Off.

void  fill::apply_rotation (char *a_shft_str, goodness g_if_fail)
{
  if (badge != Badge_Id) abort ("Fill-Badge");

  schedule *z = emp->schd;  // This schedule.

  char shft_str[21];
  strcpy (shft_str, a_shft_str);

  char *s = shft_str;
  while (*s) {
    if (*s >= 'a' && *s <= 'z') *s -= ('a' - 'A');
    s++;
  }

  goodness g_rot = Neutral;

  if (!strcmp(shft_str,"ON")) {          // Emp must work this day.

    if (day_off) {
      g_rot = g_if_fail;
    } else {
      g_rot = emp_day->g_day * -1 + Ok;
    }

  } else if (!strcmp(shft_str,"OFF")) {  // Emp must get this day off.

    if (day_off) {
      g_rot = Ok;
    } else {
      g_rot = emp_day->g_day * -1 + g_if_fail;
    }

  } else {                           // Emp must work particular shft.

    short int ishft = z->shft_abbr_to_index (shft_str);
    if (ishft < 0) {
      wlog3 ("Bad Shft \"", a_shft_str, "\" in rotation for ");
      wlog3 (emp->name, " on ", day_of_week(z->i_dow_start,iday));
      wlog  ("\nMust be \"On\", \"Off\" or a shift-name.\n");
      abort ("Abort");
    } 
    if (day_off) {
      g_rot = g_if_fail;
    } else if (shft->ishft == ishft) {
      g_rot = emp_day->g_day * -1 + Ok;
    } else {
      g_rot = emp_day->g_day * -1 + g_if_fail;
    }

  }

  if (g_rot != Neutral) {
    g_rotation += g_rot;
    g_static   += g_rot;
    if (Log_Rotations) {
      wlog2 ("      Rotation: ", emp->name);
      if (day_off) {
        wlog2 (" ", " Off");
      } else {
        wlog2 (" ", shft->name);
      }
      wlog2 (" on ", day_of_wk(Week_Start,iday));
      wlog ("  g_rot: ");
      wlogl (g_rot);
      wlog ("\n");
    }
  }

}

//------------------------------------------------------------------
