//  Brian Marshall      bmarshal@agt.net
//......................................

//  Package of functions connected with i/o.

#include "workforc.h"

static FILE *log_fp;

static char *dowa[7] = {"Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"};
static char *dow [7] = {"Monday", "Tuesday", "Wednesday",
                        "Thursday", "Friday", "Saturday", "Sunday"};

static int echo_log_flag = False;

//----------------------------------------------------------
//  Open the log file LOG.WF.  Return True on success.

int  open_log (void)
{

  log_fp = fopen ("LOG.WF", "w");

  if (log_fp) return 1;

  return 0;

}

//----------------------------------------------------------
//  Set flag indicating whether log should be echoed to cout.

void  set_echo_log (int iflag)
{
  echo_log_flag = iflag;
}

//----------------------------------------------------------
//  Write the passed string to the log file.

void  wlog (char *str)
{
  if (echo_log_flag) cout << str;

  if ( ! log_fp ) abort ("IO-2");
  if (strlen(str) > 1000) abort ("IO-1");

  fputs (str, log_fp);
  if (ferror(log_fp)) abort ("IO-3");

  fflush (log_fp);
  if (ferror(log_fp)) abort ("IO-4");
}

//----------------------------------------------------------
//  Write the passed 2 strings to the log file.

void  wlog2 (char *str1, char *str2)
{
  wlog (str1);
  wlog (str2);
}

//----------------------------------------------------------
//  Write the passed 3 strings to the log file.

void  wlog3 (char *str1, char *str2, char *str3)
{
  wlog (str1);
  wlog (str2);
  wlog (str3);
}

//----------------------------------------------------------
//  Write the passed int to the log file.

void  wlogi (int num)
{
  char  buf[30];
  sprintf (buf, "%d", num);
  wlog (buf);
}

//----------------------------------------------------------
//  Write the passed long int to the log file.

void  wlogl (long int num)
{
  char  buf[30];
  sprintf (buf, "%ld", num);
  wlog (buf);
}

//----------------------------------------------------------
//  Return the log file pointer.

FILE  *log_file_ptr (void)
{
  return log_fp;
}

//----------------------------------------------------------
//  Write the passed message to the log and abort.

void  abort (char *msg)
{
  wlog3 ("The following internal error occurred: \n   ",
         msg, "\nAbort.\n");
  wfexit (1);
}

//----------------------------------------------------------
//  Get a  long int  from the next bar delimited field
//  from the input string stream.  Abort program on problem.

long int  get_l (istrstream *istr)
{
  static char  fld[81];
  istr->getline (fld, 80, '|');
  if (istr->bad()) {
    wlog ("Error getting field from string stream.  Abort.\n");
    wfexit (1);
  }

  istrstream  fstr(fld,strlen(fld));
  if (!fstr) {
    wlog3 ("Error creating strstream from field: ", fld, "\n  ABORT.\n");
    wfexit(1);
  }

  long int l;
  fstr >> l;
  if (fstr.bad()) {
    wlog3 ("Error interpreting ", fld, "  as a number.  Abort.\n");
    wfexit (1);
  }

  return l;

}

//----------------------------------------------------------
//  Get a  short int  from the next bar delimited field
//  from the input string stream.  Abort program on problem.

short int  get_i (istrstream *istr)
{
  static char  fld[81];
  istr->getline (fld, 80, '|');
  if (istr->bad()) {
    wlog ("Error getting field from string stream.  Abort.\n");
    wfexit (1);
  }

  istrstream  fstr(fld,strlen(fld));
  if (!fstr) {
    wlog3 ("Error creating strstream from field: ", fld, "\n  ABORT.\n");
    wfexit(1);
  }

  short int i;
  fstr >> i;
  if (fstr.bad()) {
    wlog3 ("Error interpreting ", fld, "  as a short int.  Abort.\n");
    wfexit (1);
  }

  return i;

}

//----------------------------------------------------------
//  Get a maximum of len characters from the next bar
//  delimited field from the input string stream.  Skip
//  leading blanks and tabs.  Skip over trailing blanks
//  and tabs in the string stream but don't include them
//  in the output string.  Return a pointer to a null
//  string on problem or empty string.

char*  get_s (istrstream *istr, int len)
{
  static char  fld[201];

  istr->getline (fld, 200, '|');
  if (istr->bad()) {
    fld[0] = '\0';
    return fld;
  }

  fld[200] = '\0';

  char *p = fld;
  while (*p == ' ' || *p == '\t') p++;

  if (*p == '\n') {
    fld[0] = '\0';
    return fld;
  }

  char *p2 = p;
  char *last_char = null;
  while (*p2) {
    if (*p2 != ' ' && *p2 != '\t') last_char = p2;
    p2++;
  }
  if (last_char) *(last_char+1) = '\0';

  if (strlen(p) > len) *(p+len) = '\0';

  return p;

}

//----------------------------------------------------------
//  Return True if the passed string is all white space or empty.

int  all_white (char *str)
{

  while (*str) {
    if (*str != ' ' && *str != '\t' && *str != '\n') return 0;
    str++;
  }

  return 1;

}

//----------------------------------------------------------
//  Convert any lower-case letters in str to upper-case.

void upcase (char *str)
{
  while (*str) {
    if (*str >= 'a' && *str <= 'z') *str -= ('a' - 'A');
    str++;
  }
}

//----------------------------------------------------------
//  Return the day of week abbreviation for day iday in the schedule.

char*  day_of_wk (short int i_dow_start, short int iday)
{

  int i = i_dow_start + iday;
  while (i > 6) i -= 7;
  if (i < 0) i = 0;

  return dowa[i];

}

//----------------------------------------------------------
//  Return the day of week name for day iday in the schedule.

char*  day_of_week (short int i_dow_start, short int iday)
{

  int i = i_dow_start + iday;
  while (i > 6) i -= 7;
  if (i < 0) i = 0;

  return dow[i];

}

//----------------------------------------------------------
//  Given a day-of-week abbreviation (Mon, Tue etc.), return
//  iday for that day in week to be scheduled.
//  Return negative number if not a proper abbreviation.

short int dow_to_iday (short int i_dow_start, char* a_dow)
{

  char  str[21];
  strcpy (str, a_dow);

  if (str[0] >= 'a' && str[0] <= 'z') str[0] -= ('a' - 'A');
  if (str[1] >= 'A' && str[1] <= 'Z') str[1] += ('a' - 'A');
  if (str[2] >= 'A' && str[2] <= 'Z') str[2] += ('a' - 'A');

  short int num = 0;
  short int idow;
  for (idow=i_dow_start; idow<7; idow++,num++) {
    if (!strcmp(str,dowa[idow])) return Week_Start + num;
  }

  for (idow=0; idow<i_dow_start; idow++,num++) {
    if (!strcmp(str,dowa[idow])) return Week_Start + num;
  }

  return -1;

}

//---------------------------------------------------------
//  Given a character string containing a floating-point amount of
//  hours, return the number of minutes.  Return a negative number
//  on problems.

int hr_minutes (char* hrs) {

  int minutes;

  char *p = hrs;

  while (*p && *p != '.') p++;
  if (*p) *(p++) = '\n';
  minutes  = atoi(hrs) * 60;
  if (minutes < 0 || minutes > 6000) return -1;

  if (*p) {
    if (*p < '0' || *p > '9') return -1;
    int i = *p -= '0';
    minutes += i * 6;    // Each tenth of an hour is six minutes.
    p++;
  }

  if (*p) {
    if (*p < '0' || *p > '9') return -1;
    int i = *p -= '0';
    minutes += ((i * 60) / 100);
    p++;
    while (*p && (*p == ' ' || *p == '0')) p++;
    if (*p) return -1;
  }

  return minutes;

}

//----------------------------------------------------------
//  Given a number of minutes, return a pointer to a static string
//  containing the number of hours expressed as a floating number with
//  up to two decimals.

char *min_hours (int minutes)
{
  static char rec[21], buf[11];

  int hrs = minutes / 60;
  sprintf (rec, "%2d", hrs);
  minutes -= (hrs*60);

  if (minutes > 0) {
    int tenths = minutes / 6;
    buf[0] = '.';
    buf[1] = '0' + tenths;
    buf[2] = '\0';
    strcat (rec, buf);
    minutes -= (tenths*6);
  }

  if (minutes > 0) {
    int hundredths = (minutes * 100) / 60;
    buf[0] = '0' + hundredths;
    buf[1] = '\0';
    strcat (rec, buf);
  }

  return rec;

}

//----------------------------------------------------------

