//  Brian Marshall      bmarshal@agt.net
//......................................

//  Functions which write information from this schedule.

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include "workforc.h"

//-------------------------------------------------------------------------
//  Write the "best so far" schedule.

void  schedule::write_bsf (void)
{

  char   filnam[201], rec[101];

  char  cweek[4];
  cweek[0] = '\0';
  sprintf (cweek, "%3.3d", week_num);

  strcpy (filnam, "SCHD_");                 // Assemble file name.
  strcat (filnam, cweek);
  strcat (filnam, ".WF");

  wlog3 ("\nWriting file ", filnam, "...\n");

  FILE *ofp = fopen (filnam, "w");
  if (!ofp) {
    wlog3 ("Error creating file ", filnam, ".  ABORT.\n");
    wfexit(1);
  }

  strcpy (rec, "!  WorkForce Staff Schedule");
  strcat (rec, "                            Week: ");
  strcat (rec, cweek);
  fputs (rec, ofp);

  strcpy (rec, "\n\n!          ");
  strcat (rec, title);
  strcat (rec, "\n\n");
  fputs (rec, ofp);

  write_schd (Week_Start, True, ofp);

  strcpy (rec, "\n\n!  Schedule was created on: ");
  time_t t = time(NULL);
  strcat (rec, ctime(&t));
  strcat (rec, "\n");
  fputs (rec, ofp);

  fclose (ofp);
}

//-------------------------------------------------------------------------
//  Write a schedule matrix beginning on day first_day to ofs.
//  If bsf_flag is True, write the best-so-far schedule; otherwise,
//  write the current schedule.

void  schedule::write_schd (int first_day, int bsf_flag, FILE *ofp)
{
  char   rec[101], buf[31];
  int    iemp, iday, minutes;
  fill      *f;
  empl_day  *ed;
  employee  *e;

  int bad_fill;
  int any_bad = False;

  strcpy (rec, "! Employee          |");
  for (iday=Week_Start; iday<Num_Days; iday++) {
    strcat (rec, "  ");
    strcat (rec, day_of_wk(i_dow_start, iday));
    strcat (rec, " |");
  }
  strcat (rec, "  hrs");
  strcat (rec, "\n");
  if (bsf_flag) strcat (rec, "\n");
  fputs (rec, ofp);

  for (iemp=0; iemp<num_emp; iemp++) {            // For each emp...
    e = emp_array + iemp;
    sprintf (rec, "%-20s|", emp_array[iemp].name);

    minutes = 0;

    for (iday=first_day; iday<first_day+7; iday++) {    // For each day...
      ed = e->day + iday;
      if (bsf_flag) {
        f = ed->bsf_fill;
        if (!f) abort ("Schd-ws1");
        bad_fill = ed->g_bsf <= Imposible;
      } else {
        if (!ed->done) break;
        f = ed->cur_fill;
        if (!f) abort ("Schd-ws2");
        bad_fill = f->g <= Imposible;
      }
      if (f->day_off) {
        if (bad_fill) {
          strcpy  (buf, "> -- <|");
          any_bad = True;
        } else {
          strcpy  (buf, "  --  |");
        }
      } else {
        if (bad_fill) {
          sprintf (buf, ">%4s<|", f->shft->name);
          any_bad = True;
        } else {
          sprintf (buf, " %4s |", f->shft->name);
        }
      }
      strcat (rec, buf);

      if (f->shft) minutes += f->shft->minutes;

    }   // End of loop through days.

    strcat (rec, "  ");
    strcat (rec, min_hours(minutes));

    strcat (rec, "\n");

    fputs (rec, ofp);

  }    // End of loop through emps.

  if (any_bad) {
    fputs ("\nPoor fills are marked with \"> <\".\n", ofp);
  }

  fflush(ofp);

}

//----------------------------------------------------------
