//  Brian Marshall      bmarshal@agt.net
//......................................

//  schedule  member functions.

//  Other schedule member functions are in schd_rd.cpp and schd_wrt.cpp.

#include "workforc.h"

//----------------------------------------------------------
//  Flags controlling how much detail gets written to the log.
//  Note that we use (num_shft) and (!num_shft) so that the compiler
//  doesn't realize that (!num_shft) will be false and then complain
//  that the code in the if block is not reachable.

                          // Log goodness of best schedule?
#define     Log_Best_G      (num_shft)
                          // Log the tries?
#define     Log_Tries       (num_shft)
                          // Log the try's schedules?
#define     Log_Try_Schds   (!num_shft)
                          // Log the individual fills?
#define     Log_Try_Fills   (!num_shft)

//----------------------------------------------------------
//  Setup at problem start time.

void  schedule::problem_setup (int a_week_num)
{

  week_num = a_week_num;

  if ( ! (shft_array = new shift[Max_Shft]) ) {
    wlog ("Error getting memory for shifts.\n");
    wfexit(1);
  }
  shft = shft_array;   // Begin list with first element in array.

  if ( ! (emp_array = new employee[Max_Emp]) ) {
    wlog ("Error getting memory for employees.\n");
    wfexit(1);
  }
  emp = emp_array;   // Begin list with first element in array.

  wlog ("   Reading Department file...\n");
  wif dept_wif ("DEPT.WF");                   // The Department.
  if ( ! dept_wif.get() ) {
    wlog ("Did not find any data record in the Dept file.  Abort.\n");
    wfexit (1);
  }
  rec_setup (dept_wif.rec);
  dept_wif.close();

  g = Ok;
  g_prev = Imposible;
  fail = null;
  fail_prev = null;
  g_bsf = Imposible;
  worst_bsf_fill = new fill;
  worst_bsf_fill->ifill = -1;
  num_shft_done = 0;
  num_emp_done  = 0;
  hope = Very_Good * 2;
  num_fail = 0;
  tries_since_bsf = 0;

  wlog ("   Reading Shift file...\n");
  wif shft_wif ("SHIFT.WF");                  // Shifts.
  num_shft = 0;
  while (shft_wif.get()) {
    if (num_shft >= Max_Shft) {
      wlog ("Too many shifts in shift file.  Abort.\n");
      wfexit (1);
    }
    shift *s = shft_array + num_shft++;
    s->badge = Badge_Id;
    s->ishft = num_shft - 1;
    s->next_shft = null;
    if (s->ishft == 0) {
      s->prev_shft = null;
    } else {
      s->prev_shft = s - 1;
      (s->prev_shft)->next_shft = s;
    }
    s->schd = this;
    s->problem_setup (shft_wif.rec);
    if (s->done) num_shft_done++;
  }
  if (num_shft == 0) {
    wlog ("Did not find any data records in the Shift file.  Abort.\n");
    wfexit (1);
  }
  shft_wif.close();

  wlog ("   Reading Employee file...\n");
  wif emp_wif ("EMP.WF");                     // Employees.
  num_emp = 0;
  while (emp_wif.get()) {
    if (num_emp >= Max_Emp) {
      wlog ("Too many employees in employee file.  Abort.\n");
      wfexit (1);
    }
    employee *e = emp_array + num_emp++;
    e->badge = Badge_Id;
    e->iemp = num_emp - 1;
    e->next_emp = null;
    if (e->iemp == 0) {
      e->prev_emp = null;
    } else {
      e->prev_emp = e - 1;
      (e->prev_emp)->next_emp = e;
    }
    e->schd = this;
    e->problem_setup (emp_wif.rec);
    if (e->done) num_emp_done++;
  }
  if (num_emp == 0) {
    wlog ("Did not find any data records in the Employee file.  Abort.\n");
    wfexit(1);
  }
  emp_wif.close();

  done = num_shft_done == num_shft &&
         num_emp_done  == num_emp;

  if (done) {
    wlog ("\n\nNo shifts need employees.  No employees need shifts.\n");
    wlog ("There is no problem to solve.\n");
    wfexit (1);
  }

  read_lweek   ();     // Last week's schedule.
  read_avail   ();     // Posible fills for week to be scheduled.
  read_daysoff ();
  read_timeoff ();
  read_rotations ();

  check_posibility();

  int len = strlen (title);
  if (len < 1 || len > 60) abort ("Schd-21");
  if (week_num < 0 || week_num > 999) abort ("Schd-1");
  if (i_dow_start < 0 || i_dow_start > 6) abort ("Schd-3");
  if (num_try < 1 || num_try > 10000) abort ("Schd-4");
  if (num_shft < 1 || num_shft > Max_Shft) abort ("Schd-5");
  if (num_emp < 1 || num_emp > Max_Emp) abort ("Schd-7");
  if (fail) abort ("Schd-12");
  if (fail_prev) abort ("Schd-13");
  if (g_bsf != Imposible) abort ("Schd-14");
  if (g < Imposible || g > Manditory) abort ("Schd-9");
  if (g_prev < Imposible || g_prev > Manditory) abort ("Schd-10");
  if (hope < Irrelevant || hope > Critical) abort ("Schd-15");
  if (num_shft_done < 0 ||
      num_shft_done > num_shft) abort ("Schd-16");
  if (num_emp_done  < 0 ||
      num_emp_done  > num_emp) abort ("Schd-17");
  if (done) abort ("Schd-18");
  if (num_fail != 0) abort ("Schd-19");

}

//----------------------------------------------------------
//  Setup the schedule given the department file record.

void  schedule::rec_setup (char *rec)
{
  char  buf[31];

  istrstream istr(rec,strlen(rec));
  if (!istr) {
    wlog ("Error creating string stream from record: \n");
    wlog2 (rec, "\n  ABORT.\n");
    wfexit(1);
  }

  strcpy (buf, get_s(&istr,20));
  upcase (buf);
  if ( ! strcmp (buf, "MON") ) {
    i_dow_start = 0;
  } else if ( ! strcmp (buf, "TUE") ) {
    i_dow_start = 1;
  } else if ( ! strcmp (buf, "WED") ) {
    i_dow_start = 2;
  } else if ( ! strcmp (buf, "THU") ) {
    i_dow_start = 3;
  } else if ( ! strcmp (buf, "FRI") ) {
    i_dow_start = 4;
  } else if ( ! strcmp (buf, "SAT") ) {
    i_dow_start = 5;
  } else if ( ! strcmp (buf, "SUN") ) {
    i_dow_start = 6;
  } else {
    wlog ("\nWeek-Start must be one of:\n");
    wlog ("   Mon, Tue, Wed, Thu, Fri, Sat or Sun.\n");
    wfexit(1);
  }

  num_try = get_i(&istr);
  if (num_try < 1 || num_try > 9999) {
    wlog ("\nNum-Try must be between 1 and 9999.\n");
    wfexit(1);
  }

  strcpy (title, get_s(&istr,60));

}

//----------------------------------------------------------
//  Look for a good schedule and write output files.  (Public)
//  Return True if successful.

int  schedule::find_good_schedule (void)
{

  for (itry=0; itry<num_try; itry++) {         //  Keep trying...

    if (Log_Tries) wlog  ("\n");
    wlog  ("--- Try: ");
    wlogi (itry+1);
    wlog  ("\n");

    try_setup();

    check_integrity();

    try_to_complete ();

    other_g_aspects();  // Seniority aspects, worst fill etc.

    if (done && g > Imposible && g > g_bsf) {   // Best so far.
      save_bsf();
      tries_since_bsf = 0;
    } else {
      tries_since_bsf++;
    }

    note_try();

    if (hope <= 0) {
      broaden_horizons();
    }

    backout();

  }   // End of loop of trys.

  if (Log_Tries) {
    set_echo_log (False);
    log_failures();
    set_echo_log (True);
  }

  if (Log_Best_G) {
    wlog  ("\nGoodness of best schedule found: ");
    wlogl (g_bsf);
    wlog ("\n");
  }

  if (g_bsf > Imposible) return True;

  return False;

}

//----------------------------------------------------------
//  Setup for before we attempt a completion.

void  schedule::try_setup (void)
{

  if (Log_Tries) {
    set_echo_log (False);
    wlog  ("         (starting goodness: ");
    wlogl (g);
    wlog  ("\n");
    set_echo_log (True);
  }

  if (itry < 7) {
    g_cutoff = Poor;       // Fills' g_static must be better than this.
  } else if (g_bsf >= Ok) {
    g_cutoff = Poor + Poor - 1;
  } else if (itry < num_try/2) {
    g_cutoff = Bad;
  } else {
    g_cutoff = Imposible;
  }

  short int ishft;
  for (ishft=0; ishft<num_shft; ishft++) {
    shift *s = shft_array + ishft;
    s->try_setup();
  }

  short int iemp;
  for (iemp=0; iemp<num_emp; iemp++) {
    employee *e = emp_array + iemp;
    e->try_setup();
  }

  worst_fill = null;

}

//----------------------------------------------------------
// Make an attempt to complete this schedule to the last day.

void schedule::try_to_complete (void)
{
  fill  *f;

  int n = 0;

  while ( ! done ) {

    if (Log_Try_Fills) {
      set_echo_log (False);
      wlog ("Pick: ");
      wlogi (++n);
      wlog  (" ");  // \n will be added by pick_fill().
      set_echo_log (True);
    }

    f = pick_fill();

    if (!f) return;    // Failed finding e for s or s for e.

    if (Log_Try_Fills) {
      set_echo_log (False);
      f->log_hope();
      set_echo_log (True);
    }

    do_fill (f);

  }

}

//----------------------------------------------------------
// Pick the most hopeful thing to do for the most important emp or shft.
// Return a pointer to the fill to do (some empl_day's cur_fill).

fill*  schedule::pick_fill (void)
{

  fill *f;

  fill *hf = null;      // Most hopeful fill found so far.

  if (emp->imp >= shft->imp) {   // Most important emp is most important.

    if (Log_Try_Fills) {
      set_echo_log (False);
      emp->log_imp();
      set_echo_log (True);
    }

    f = emp->cur_day->pos_fill;
    for (; f && f->hope>Imposible; f=f->next_fill) {
      if (f->g_static <= g_cutoff) continue;
      if (f->day_off) {
        hf = f;
        break;              // Use first found that is a day off...
      }
      shft_day *sd = f->shft->day + f->iday;
      if (sd->emps_have < sd->emps_req) {
        hf = f;
        break;              // or for a shft that needs more emps.
      }
    }

    if (!hf) {
      if (!fail) fail = new failure;
      if (!fail) abort ("PFill-1");
      fail->iday = emp->cur_day->iday;
      fail->emp  = emp;
      fail->shft = null;
      if (Log_Tries) {
        set_echo_log (False);
        wlog2 ("\n  -- Failed to find shft for emp: ", emp->name);
        wlog  ("  on day ");
        wlogi (emp->cur_day->iday);
        wlog ("\n");
        set_echo_log (True);
      }
    }

  } else {         // Most important shft is most important.

    if (Log_Try_Fills) {
      set_echo_log (False);
      shft->log_imp();
      set_echo_log (True);
    }

    employee *e;;
    for (e=emp; e; e=e->next_emp) {
      if (e->cur_day->iday != shft->cur_day->iday) continue;
      if (e->cur_day->cur_fill) continue;
      for (f=e->cur_day->pos_fill; f; f=f->next_fill) {
        if (f->shft && f->shft->ishft == shft->ishft) {
          if (f->g_static <= g_cutoff) continue;
          if (!hf || f->hope+f->g > hf->hope+hf->g) hf = f;
//          if (!hf || f->g > hf->g) hf = f;
        }
      }  // End of loop through fills.
    }     // End of loop through emps.

    if (!hf) {
      if (!fail) fail = new failure;
      if (!fail) abort ("PFill-2");
      fail->iday = shft->cur_day->iday;
      fail->emp  = null;
      fail->shft = shft;
      if (Log_Tries) {
        set_echo_log (False);
        wlog2 ("\n  -- Failed to find emp for shft: ", shft->name);
        wlog  ("  on day ");
        wlogi (shft->cur_day->iday);
        wlog  ("\n");
        set_echo_log (True);
      }
    }

  }  // End of block for most important shift.

  return hf;

}

//----------------------------------------------------------
//  Apply this fill to the schedule.

void  schedule::do_fill (fill *f)
{

  if (f->emp->done) abort ("Schd-df2");
  if (f->shft && f->shft->done) abort ("Schd-df1");

  if (f->shft) {
    f->shft->do_fill(f);
    if (f->shft->done) {
      num_shft_done++;
      if (Log_Try_Fills) {
        set_echo_log (False);
        wlog ("  Shift Done.  num_shft_done: ");
        wlogi (int(num_shft_done));
        wlog ("\n");
        set_echo_log (True);
      }
    }
  }

  f->emp->do_fill(f);
  if (f->emp->done) {
    num_emp_done++;
    if (Log_Try_Fills) {
      set_echo_log (False);
      wlog ("  Employee Done.  num_emp_done: ");
      wlogi (int(num_emp_done));
      wlog ("\n");
      set_echo_log (True);
    }
  }

  if (num_shft_done == num_shft &&
      num_emp_done  == num_emp) done = True;

  g += f->g;

  if (!worst_fill || f->g < worst_fill->g) worst_fill = f;

  if (f->shft) {
    short int ie;
    for (ie=0; ie<num_emp; ie++) {
      employee *e = emp_array + ie;
      empl_day *ed = e->day + f->iday;
      if (ed->done) continue;
        fill *pf = ed->pos_fill;
        while (pf) {
          if (pf->shft && pf->shft->ishft == f->shft->ishft) {
            e->det_imp();     // Re-evaluate emp's importance.
            break;
          }   // End of loop through emp's posible fills.
          pf = pf->next_fill;
        }   // End of block for employee-days which are not done.
      }   // End of loop through employees.
    }


}

//----------------------------------------------------------
//  Un-apply this fill to the schedule.

void  schedule::undo_fill (fill *f)
{

  if (f->emp->done) num_emp_done--;
  f->emp->undo_fill(f);

  if (f->shft) {
    if (f->shft->done) num_shft_done--;
    f->shft->undo_fill(f);
  }

  done = False;

  g -= f->g;

}

//----------------------------------------------------------
//  Remember the current schedule as the "best so far" schedule.

void  schedule::save_bsf (void)
{

  employee *e = emp;
  while (e) {   // For each emp...
    e->save_bsf();
    e = e->next_emp;
  }

  g_bsf = g;

  *worst_bsf_fill = *worst_fill;

  if (Log_Tries) {
    set_echo_log (False);
    wlog ("*** Best So Far ***\n");
    set_echo_log (True);
  }

  if (Log_Best_G) {
    wlog  ("   Best So Far...   Goodness: ");
    wlogl (g);
    wlog  ("\n");
  }

}

//----------------------------------------------------------
//  I have little hope for improvement; I must change the way I look at
//  things so as to try different things.

void  schedule::broaden_horizons (void)
{

  if (Log_Tries) {
    set_echo_log (False);
    wlog ("\n### Broadening Horizons...\n");
    set_echo_log (True);
  }

  short int  i;

  shift  *s;
  for (i=0; i<num_shft; i++) { 
    s = shft_array + i;
    s->broaden_horizons();
  }

  employee  *e;
  for (i=0; i<num_emp; i++) { 
    e = emp_array + i;
    e->broaden_horizons();
  }

  hope = Very_Good * 2;
  tries_since_bsf = 0;

}

//----------------------------------------------------------
//  Check that it is posible to satisfy the shifts with the available emps.
//  If it is not, print a message and abort.
//  We want to establish that:
//    - For each day:
//      - For each shft:
//        - Enough emps that can work shft
//    - For the week, the sum of days_per_week of the emps is not less
//      than the sum of the emps_req of the shfts.
//    - For the week, the sum of min_per_week of the emps is not less
//      than the sum of the minutes*emps_req of the shifts.
//  We should check the minutes each shft requires and the minutes of each
//  emp that can work it.

void  schedule::check_posibility (void)
{
  int    iday, iemp, ishft, ifill;
  int    emps_req_day_shft, emps_have_day_shft;
  int    emps_req_week, emps_have_week;
  int    emp_minutes, shft_minutes;
  fill  *f;

  emps_req_week = 0;
  shft_minutes = 0;

  for (iday=Week_Start; iday<Num_Days; iday++) {

    for (ishft=0; ishft<num_shft; ishft++) {

      emps_req_day_shft = shft_array[ishft].day[iday].emps_req;
      emps_req_week += emps_req_day_shft;
      shft_minutes += (emps_req_day_shft * shft_array[ishft].minutes);
      emps_have_day_shft = 0;
      for (iemp=0; iemp<num_emp; iemp++) {
        empl_day *ed = emp_array[iemp].day + iday;
        int nfill = ed->num_pos_fill;
        for (ifill=0; ifill<nfill; ifill++) {
          f = ed->pos_fill_array + ifill;
          if (f->shft && f->shft->ishft == ishft) {
            emps_have_day_shft++;
            break;
          }
        }
      }

      if (emps_have_day_shft < emps_req_day_shft) {
         wlog  ("\nNot enough employees for shift ");
         wlog  (shft_array[ishft].name);
         wlog2 (" for ", day_of_week(i_dow_start,iday));
         wlog  (".  Abort.\n");
         wfexit(1);
      }

    }    // End of loop through shifts.

  }    // End of loop through days.

  short int check_minutes = True;

  emps_have_week = 0;
  emp_minutes = 0;
  for (iemp=0; iemp<num_emp; iemp++) {
    emps_have_week += emp_array[iemp].days_per_week;
    emp_minutes += emp_array[iemp].min_per_week;
    if (emp_array[iemp].min_per_week == 0) check_minutes = False;
  }

  if (emps_have_week < emps_req_week) {
    wlog2 ("\nWarning... Not enough employee days for shift days.\n",
           "\nThis schedule will require over-time.  Continuing anyway...\n");
  }

  if (check_minutes && (shft_minutes - emp_minutes) > 60) {
    wlog  ("\nWarning... Not enough employee hours for shift hours.\n");
    wlog  ("           Approximate shortage: ");
    wlogi ((shft_minutes-emp_minutes)/60);
    wlog  ("           hours.  This schedule will require over-time.");
    wlog  ("           Continuing anyway...\n");
  }

}

//----------------------------------------------------------
// Having just attempted to complete the schedule, backout each emp
// to i_week_start.

//  Later: Don't have to back out each emp all the way back to i_w_s;
//         backout untill what remains has enough hope?

//  NOTE: the management of sd.emps_can and sd.emps_could by calls
//        to sd.note_can(), sd.note_could() and sd.note_busy by
//        various ed functions may require backout all the way to iws.

void schedule::backout (void)
{
  short int  iday, iemp;

  for (iemp=0; iemp<num_emp; iemp++) {
    employee *e = emp_array + iemp;

    for (iday=e->cur_day->iday; iday>=Week_Start; iday--) {
      empl_day *ed = e->day + iday;

      if (ed->done) {
        fill *cf = ed->cur_fill;
        undo_fill (cf);
      }

    }   // End of loop back through days.

  }   // End of loop through emps.

  g -= g_seniority;
  if (worst_fill) g -= (worst_fill->g * 2);

}

//----------------------------------------------------------
//  Note that an attempt has been made to complete the schedule.
//  This function causes tries to differ.

void  schedule::note_try (void)
{

  short int ishft;
  for (ishft=0; ishft<num_shft; ishft++) {
    shift *s = shft_array + ishft;
    s->note_try();
  }

  short int iemp;
  for (iemp=0; iemp<num_emp; iemp++) {
    employee *e = emp_array + iemp;
    e->note_try();
  }

  if (done) {                     // Try was sucessfull...
    if (g > g_bsf) {
      g_try = Very_Good;
    } else if (g_bsf > Poor) {
      if (g > Poor) {
        if (g > g_prev) {
          g_try = Good;
        } else {
          g_try = Ok;
        }
      } else {
        g_try = Bad;
      }
    } else {                  // No best-so-far or worse than Poor.
      if (g > Poor) {
        if (g > g_prev) {
          g_try = Good;
        } else {
          g_try = Ok;
        }
      } else {
        if (g > g_prev) {
          g_try = Ok;
        } else {
          g_try = Poor;
        }
      }
    }
  } else {                        // Try failed...
    num_fail++;
    if (fail_prev) {                 // and previous try also failed...
      if (fail->iday > fail_prev->iday) {
        g_try = Good;
      } else if (fail->emp  != fail_prev->emp ||
                 fail->shft != fail_prev->shft) {
        g_try = Ok;
      } else {
        g_try = Poor;
      }
    } else {                         // and previous try was sucessfull...
      g_try = Poor;
    }
  }

  if (g_try >= Good) {
    hope += 2;
  } else if (tries_since_bsf > num_try/5) {
    hope = 0;
  } else if (g_try > Ok) {
    hope++;
  } else if (g_try <= Bad) {
    hope -= 2;
  } else if (g_try <= Poorish) {
    hope -= 2;
  } else {
    hope -= 1;
  }

  g_prev = g;

  if (fail_prev) delete fail_prev;
  fail_prev = fail;
  fail = null;

  if (Log_Try_Schds) {
    set_echo_log (False);
    write_schd (Week_Start, False, log_file_ptr());
    set_echo_log (True);
  }

  if (Log_Tries) {
    set_echo_log (False);
    wlog  ("   Goodness: ");
    wlogl (g);
    wlog  ("   Sen-Fctr: ");
    wlogl (g_seniority);
    wlog  ("   Try-Success: ");
    wlogl (g_try);
    wlog  ("   Hope: ");
    wlogl (hope);
    wlog  ("\n");
    set_echo_log (True);
  }

}

//---------------------------------------------------------
//  Given a shift abbreviation (which may contain leading and/or
//  trailing blanks), return the index of the shft in the shft_array.
//  Return -1 if sname is "--" (a day off).  Return -9 on error.

int  schedule::shft_abbr_to_index (char *sname)
{

  char *p, buf[31];
  int   ishft;

  strcpy (buf, sname);

  p = buf;
  while (*p) p++;
  while (p > buf) {
    p--;
    if (*p != ' ') break;
  }
  *(++p) = '\0';        // Remove trailing blanks.

  p = buf;
  while (*p && *p == ' ') p++;  // Skip past leading blanks.

  if ( strcmp (p, "--") == 0 ) return -1;

  for (ishft=0; ishft<num_shft; ishft++) {
    if ( strcmp (p, shft_array[ishft].name) == 0 ) return ishft;
  }

  return -9;

}

//---------------------------------------------------------
//  Given an employee name (which may contain leading and/or
//  trailing blanks), return the index of the employee in the
//  emp_array.  Return -9 on error.

int  schedule::emp_name_to_index (char *ename)
{

  int iemp;
  for (iemp=0; iemp<num_emp; iemp++) {
    if ( strcmp (ename, emp_array[iemp].name) == 0 ) return iemp;
  }

  return -9;

}

//----------------------------------------------------------
//  Given an employee id, return the index of the emp in the emp_array.
//  Return -9 on error.

int  schedule::emp_id_to_index (char *id)
{

  int   iemp;
  for (iemp=0; iemp<num_emp; iemp++) {
    if (strcmp(emp_array[iemp].id,id) == 0) return iemp;
  }

  return -9;

}

//----------------------------------------------------------
//  Write information about failed tries to the log.

void  schedule::log_failures (void)
{
  wlog ("\n\n----------------------------------------------------\n\n");
  wlog ("                 Failed-Try Statistics                  \n\n");

  wlog  ("\nNumber of failed tries: ");
  wlogi (int(num_fail));
  wlog  ("\n");

  int  ishft, iemp, iday;
  char  rec[200], buf[30];
  empl_day *ed;
  shft_day *sd;

  wlog ("\nFailures finding an employee for a shift:\n");
  wlog ("         Number of times involved in a failure:\n\n");

  strcpy (rec, "  Shift |");
  for (iday=Week_Start; iday<Num_Days; iday++) {
    strcat (rec, "  ");
    strcat (rec, day_of_wk(i_dow_start, iday));
    strcat (rec, " |");
  }
  strcat (rec,"\n");
  wlog (rec);

  for (ishft=0; ishft<num_shft; ishft++) {            // For each shft...

    if (shft_array[ishft].num_fail <= 0) continue;

    sprintf (rec, "   %-4s |", shft_array[ishft].name);

    for (iday=Week_Start; iday<Num_Days; iday++) {    // For each day...
      sd = shft_array[ishft].day + iday;
      sprintf (buf, " %4d |", int(sd->num_fail));
      strcat (rec, buf);
    }   // End of loop through days.

    strcat (rec,"\n");
    wlog (rec);

  }    // End of loop through shfts.

  wlog ("\nFailures finding a shift for an employee:\n");
  wlog ("         Number of times involved in a failure:\n\n");

  strcpy (rec, "      Employee         |");
  for (iday=Week_Start; iday<Num_Days; iday++) {
    strcat (rec, "  ");
    strcat (rec, day_of_wk(i_dow_start, iday));
    strcat (rec, " |");
  }
  strcat (rec,"\n");
  wlog (rec);

  for (iemp=0; iemp<num_emp; iemp++) {            // For each emp...

    if (emp_array[iemp].num_fail <= 0) continue;

    sprintf (rec, "   %-20s|", emp_array[iemp].name);

    for (iday=Week_Start; iday<Num_Days; iday++) {    // For each day...
      ed = emp_array[iemp].day + iday;
      sprintf (buf, " %4d |", int(ed->num_fail));
      strcat (rec, buf);
    }

    strcat (rec,"\n");
    wlog (rec);

  }    // End of loop through emps.

  wlog ("\n\n----------------------------------------------------\n\n");

}

//----------------------------------------------------------
//  Write information about the worst fill in the best-so-far
//  schedule to the log.

void  schedule::log_worst_bsf_fill (void)
{

  fill *f = worst_bsf_fill;
  if (!f) return;

  wlog  ("\n\n      ----------------------------------------\n\n");
  wlog  ("                  The Worst Aspect                \n\n");

  if (f->g >= Ok) {
    wlog ("Workforce does not think any single aspect of the\n");
    wlog ("best schedule is worse than \"Ok\".\n");
    return;
  }

  wlog  ("In the best schedule found, WorkForce considers\n");
  wlog  ("the worst aspect to be:\n\n");

  if (f->day_off && f->g_work_today < Ok &&
      f->g_static >= Ok && f->g_shft_chg >= Ok && f->g_mpw >= Ok) {
    wlog3 ("  By ", day_of_week (Week_Start, f->iday), ", ");
    wlog2 (f->emp->name, " has too many days off.\n");
    wlog3 ("\nWorkForce considers this to be ", g2str(f->g), ".\n");
    return;
  }

  wlog2 ("  Giving ", f->emp->name);   

  if (f->day_off) {
    wlog (" the day off on ");
  } else {
    wlog3 (" the ", f->shft->name, " shift on ");
  }

  wlog2 (day_of_week (Week_Start, f->iday), ".\n");

  wlog2 ("\nWorkForce considers this to be ", g2str(f->g));

  wlog (" because:\n");

  if (f->g_static < Ok) {
    wlog2 ("  o Availability this shift on this day is ",
           g2str(f->g_static));
    wlog (".\n");
  }

  if (f->g_shft_chg < Ok) {
    wlog2 ("  o Shift-change aspects are ", g2str(f->g_shft_chg));
    wlog (".\n");
  }

  if (f->g_work_today < Ok) {
    if (f->day_off) {
      wlog2 ("  o Getting the day off is ", g2str(f->g_work_today));
    } else {
      wlog2 ("  o Working this day is ", g2str(f->g_work_today));
    }
    wlog (".\n");
  }

  if (f->g_mpw < Ok) {
    wlog2 ("  o Hours/week aspects are ", g2str(f->g_mpw));
    wlog (".\n");
  }

}

//----------------------------------------------------------
//  Convert a user-goodness code to a goodness.
//  Return True on success.

int schedule::ug2g (char *ug, goodness *g)
{
  char *p = ug;
  if (*p >= 'a' && *p <= 'z') *p -= ('a' - 'A');

  if (*p == 'N') {        // No         Don't ever do this.
    *g = Imposible;
  } else if (*p == 'V') { // Very-bad   Do only if absolutly necessary.
    *g = Very_Bad;
  } else if (*p == 'B') { // Bad        Try not to do this.
    *g = Bad;
  } else if (*p == 'P') { // Poor       Do only if situation requires it.
    *g = Poor;
  } else if (*p == 'E') { // Expected   Expected (but not good).
    *g = Expected;
  } else if (*p == 'O') { // OK         Normal thing to do.
    *g = Ok;
  } else if (*p == 'G') { // Good       Preferred thing to do.
    *g = Good;
  } else if (*p == '-') { // Neutral    Neutral or no effect.
    *g = Neutral;
  } else {
    return False;
  }

  return True;

}

//----------------------------------------------------------
//  Given a goodness, return a pointer to a static string containing
//  a phrase describing the goodness.

char *schedule::g2str (goodness g)
{
  static char very_bad[]  = "Very-Bad";
  static char bad[]       = "Bad";
  static char poor[]      = "Poor";
  static char poorish[]   = "Poorish";
  static char ok[]        = "Ok";
  static char good[]      = "Good";

  if (g > 2) {
    return good;
  } else if (g >= Ok) {
    return ok;
  } else if (g > -5) {
    return poorish;
  } else if (g > -20) {
    return poor;
  } else if (g > -100) {
    return bad;
  } else {
    return very_bad;
  }

}

//----------------------------------------------------------
//  Check data-structure integrity.

void  schedule::check_integrity (void)
{

  shift *s = shft;
  int num_s = 0;
  while (s) {
    if (++num_s > num_shft) abort ("Schd-20");
    if (shft_array[s->ishft].ishft != s->ishft) abort ("Schd-22");
    s->check_integrity();
    s = s->next_shft;
    if (s) {
      if (s->prev_shft->next_shft->ishft != s->ishft) abort ("Schd-24");
    }
  }
  if (num_s != num_shft) abort ("Schd-26");

  employee *e = emp;
  int num_e = 0;
  while (e) {
    if (++num_e > num_emp) abort ("Schd-21");
    if (emp_array[e->iemp].iemp != e->iemp) abort ("Schd-23");
    e->check_integrity();
    e = e->next_emp;
    if (e) {
      if (e->prev_emp->next_emp->iemp != e->iemp) abort ("Schd-25");
    }
  }
  if (num_e != num_emp) abort ("Schd-27");

}

//----------------------------------------------------------
//  Determine the goodness associated with how the goodness/baddness
//  is distributed among the employees.  The ideal is that employees
//  of the same seniority are all the same goodness and that no
//  employee does worse than someone of lower seniority.

void  schedule::det_g_seniority (void)
{
  g_seniority = Ok;

  int i, iemp;
  short int min_sen;

  short int sen[Max_Emp];            // List of employee seniorities.
  for (i=0; i<num_emp; i++) sen[i] = emp_array[i].seniority;

  goodness   g_this, g_diff, tmp;
  short int  sen_this;

  goodness   g_previous = Imposible;
  short int  sen_prev = 999;

  while (True) {                 // For each employee...
    min_sen = 9999;              // in order of increasing seniority...
    iemp = -1;
    for (i=0; i<num_emp; i++) {
      if (sen[i] < min_sen) {
        min_sen = sen[i];
        iemp = i;
      }
    }
    if (min_sen > 99) break;

    g_this = emp_array[iemp].g;
    sen_this = sen[iemp];
    sen[iemp] = 999;      // Don't pick this one again.

    if (sen_prev <= 99) {       // Don't do this for first one.

      g_diff = g_this - g_previous;

      if (g_diff < 0 && sen_this > sen_prev) {

        tmp = g_diff + (sen_prev - sen_this);
        tmp = (tmp * sen_this) / Imp_Divsr;
        g_seniority += (tmp / 4);             // Will make more negative.

      } else if (g_diff > 0 && sen_this == sen_prev) {

        g_seniority -= 4;

      }

    }   // End of block done on all but first one.

    g_previous = g_this;
    sen_prev = sen_this;

  }  //  End of loop through emps in order of increasing seniority.

}

//----------------------------------------------------------
//  Consider aspects of the schedule other than the sum of fill
//  goodnesses, and apply them to the schedule goodness.

void  schedule::other_g_aspects (void)
{
                        // Give more weight to worst fill.
  if (worst_fill) g += (worst_fill->g * 2);

  det_g_seniority ();  // Determine seniority-order goodness factor.
  g += g_seniority;

}

//----------------------------------------------------------
