//  Brian Marshall      bmarshal@agt.net
//......................................

//  Shift functions.

#include <stdlib.h>
#include <string.h>
#include "workforc.h"


//------------------------------------------------------------------------
//  Setup at problem-startup using a record from the shft file.

void  shift::problem_setup (char *rec)
{
  if (badge != Badge_Id) abort ("Shft-ps1");

  if ( ! (day = new shft_day[Num_Days]) ) {
    wlog ("Error getting memory for shift-days.\n");
    wfexit(1);
  }

  rec_setup (rec);

  i_been_prob = Trivial;
  cur_day = day + Week_Start;
  done = False;
  num_fail = 0;
  imp = -999999;

  while (cur_day->done) {
    if (cur_day->next_day) {
      cur_day++;
    } else {
      done = True;
      break;
    }
  }

  int len = strlen (name);
  if (len < 1 || len > 4) abort ("Shft-20");
  if (start_time < 0 || start_time > 2400) abort ("Shft-2");
  if (end_time   < 0 || end_time   > 2400) abort ("Shft-3");
  if (i_been_prob != Trivial) abort ("Shft-5");
  if (num_fail) abort ("Shft-8");

}

//------------------------------------------------------------------------
//  Apply the information in the record from the shift file.

void  shift::rec_setup (char *rec)
{
  if (badge != Badge_Id) abort ("Shft-rs1");

  istrstream istr(rec,strlen(rec));     // Make a string-stream.
  if (!istr) {
    wlog ("Error creating strstream from record: \n");
    wlog2 (rec, "\n  ABORT.\n");
    wfexit(1);
  }

  strcpy (name, get_s(&istr,4));
  start_time = get_i(&istr);
  end_time   = get_i(&istr);

  char buf[21];
  strcpy (buf, get_s(&istr,20));
  minutes = hr_minutes (buf);
  if (minutes < 6 || minutes > 720) {
    wlog3 ("Incorrect Hours  \"", buf, "\" on line: \n");
    wlog2 (rec, "\n  ABORT.\n");
    wfexit(1);
  }

  short int iday;
  for (iday=0; iday<Num_Days; iday++) {
    shft_day *sd = day + iday;
    sd->badge = Badge_Id;
    sd->iday = iday;
    if (iday == Num_Days-1) {
      sd->next_day = null;
    } else {
      sd->next_day = sd + 1;
    }
    if (iday == 0) {
      sd->prev_day = null;
    } else {
      sd->prev_day = sd - 1;
    }
    sd->shft = this;
    short int emps_req;
    if (iday < Week_Start) {
      emps_req = get_i (&istr);
    } else {
      emps_req = (sd-Week_Len)->emps_req;
    }
    sd->problem_setup (emps_req);
  }

}

//------------------------------------------------------------------------
//  Setup to attempt a completion.

void  shift::try_setup (void)
{
  if (badge != Badge_Id) abort ("Shft-ts1");

  int iday;
  for (iday=cur_day->iday; iday<Num_Days; iday++) {   // Current week.
    shft_day *d = day + iday;
    d->try_setup();
  }

  while (cur_day->done) {
    if (cur_day->iday == Num_Days-1) {
      done = True;
      break;
    } else {
      cur_day++;
    }
  }

}

//------------------------------------------------------------------------
//  Apply the passed fill to this shift.

void  shift::do_fill (fill* f)
{
  if (badge != Badge_Id) abort ("Shft-df1");

  if (!f->shft || f->shft != this) {
    wlog ("Internal Error: Bad shift in shift::do_fill().  Abort.\n");
    wfexit(1);
  }

  shft_day *sd = day + f->iday;

  sd->do_fill(f);

  while (cur_day->done) {
    if (cur_day->iday == Num_Days-1) {
      done = True;
      det_imp();
      break;
    } else {
      cur_day++;
      det_imp();
    }
  }

}
//------------------------------------------------------------------------
//  Un-apply the passed fill to this shift.

void  shift::undo_fill (fill* f)
{
  if (badge != Badge_Id) abort ("Shft-uf1");

  if (!f->shft || f->shft != this) {
    wlog ("Internal Error: Bad shift in shift::undo_fill().  Abort.\n");
    wfexit(1);
  }

  shft_day *sd = day + f->iday;

  sd->undo_fill (f);

  if (cur_day->iday > f->iday) cur_day = day + f->iday;

  done = False;

}

//------------------------------------------------------------------------
//  Determine the importance of this shft, set it in the
//  data, and, if the shft is part of a schedule, move the
//  shft in the list so as to keep the list in decreasing
//  order of importance.

void  shift::det_imp (void)
{
  if (badge != Badge_Id) abort ("Shft-di1");

  if (done) {
    imp = -999999;
  } else {
    short int tot = cur_day->emps_could + cur_day->emps_can;
    tot += cur_day->emps_have;
    if (tot <= cur_day->emps_req &&
        (cur_day->emps_can > 0 || cur_day->emps_could == 0)) {
      imp = 999999;
    } else {
      imp = -99999;
    }
  }

#define  prev_OLElm  prev_shft
#define  next_OLElm  next_shft
#define  OLKey       imp
#define  OLHead      (schd->shft)
#define  OLElmTyp    shift

  if (schd) {                       // Ensure that most important is first.
    #include "olsetkey.inc"
  }

#undef  prev_OLElm
#undef  next_OLElm
#undef  OLKey
#undef  OLHead
#undef  OLElmTyp

}

//----------------------------------------------------------
//  A try involving this shift has finished.

void  shift::note_try (void)
{
  if (badge != Badge_Id) abort ("Shft-nt1");

  short int iday;
  for (iday=Week_Start; iday<Num_Days; iday++) {
    shft_day *sd = day + iday;
    sd->note_try();
  }

  if (schd->fail) {
    if (schd->fail->shft == this) {
      num_fail++;
      i_been_prob += 128;
    }
  } else {
    if (schd->g > Imposible && i_been_prob > 16) {
      if (i_been_prob > 32) {
        i_been_prob = 32;
      } else {
        i_been_prob -= 4;
      }
    }
  }

}

//------------------------------------------------------------------------
//  Check the data-structure integrity.

void  shift::check_integrity (void)
{
  if (badge != Badge_Id) abort ("Shft-Badge");

  shft_day *d = day;
  int num_d = 0;
  while (d) {
    if (++num_d > Num_Days) abort ("Shft-9");
    d->check_integrity();
    d = d->next_day;
    if (d) {
      if (d->prev_day->next_day->iday != d->iday) abort ("Shft-10");
    }
  }
  if (num_d != Num_Days) abort ("Shft-18");

}

//------------------------------------------------------------------------
//  Write a line to the log identifying this shift and indicating
//  why it is important.

void  shift::log_imp (void)
{
  if (badge != Badge_Id) abort ("Shft-Badge");

  char str[101];

  sprintf (str, "S%2d: %2dD %7ldI | %2ldbp %2lddbp\n",
           int(ishft), int(cur_day->iday), imp, i_been_prob,
           cur_day->i_been_prob);

  wlog (str);

}

//----------------------------------------------------------
//  I have little hope for improvement; I must change the way I look at
//  things so as to try different things.

void  shift::broaden_horizons (void)
{
  if (badge != Badge_Id) abort ("Shft-bh1");

  int iday;
  for (iday=Week_Start; iday<Num_Days; iday++) {
    shft_day *sd = day + iday;
    sd->broaden_horizons();
  }

  i_been_prob /= 2;

}

//------------------------------------------------------------------------
