//  Brian Marshall      bmarshal@agt.net
//......................................

//  shft_day  member functions.

#include <stdlib.h>
#include "workforc.h"

//----------------------------------------------------------
//  Setup to find a good schedule.

void  shft_day::problem_setup (short int a_emps_req)
{
  if (badge != Badge_Id) abort ("Sday-ps1");

  emps_req = a_emps_req;

  if (iday < Week_Start || emps_req == 0) {
    emps_have = emps_req;
    done = True;
  } else {
    emps_have = 0;
    done = False;
  }

  i_been_prob = Trivial;

  num_fail = 0;

  if (i_been_prob != Trivial) abort ("Sday-4");
  if (num_fail) abort ("Sday-7");
}

//------------------------------------------------------------------
// Setup before we can attempt a completion.

void  shft_day::try_setup (void)
{
  if (badge != Badge_Id) abort ("Sday-ts1");

  emps_could = emps_can = 0;

// This may mean schd->backout() has to back all the way out.

}

//----------------------------------------------------------
//  Apply the fill to this shft_day.

void  shft_day::do_fill (fill *f)
{
  if (badge != Badge_Id) abort ("Sday-df1");
  if (f->shft != shft || f->iday != iday) abort ("Sday-df2");

  emps_have++;

  if (emps_have > emps_req) abort ("Sday-df3");

  if (emps_have == emps_req) done = True;

}

//----------------------------------------------------------
//  Un-apply the fill to this shft_day.

void  shft_day::undo_fill (fill *f)
{
  if (badge != Badge_Id) abort ("Sday-uf1");

  if (!f->shft) {
    wlog ("Internal Error: Bad call to shft_day::undo_fill().  Abort.\n");
    wfexit(1);
  }

  emps_have--;

  if (emps_have < 0) {
    wlog ("Internal Error: Undo too many emps for shift-day.  Abort.\n");
    wfexit(1);
  }

  done = False;

}

//----------------------------------------------------------
//  An try involving this shft_day has finished.

void  shft_day::note_try (void)
{
  if (badge != Badge_Id) abort ("Sday-nt1");

  schedule *z = shft->schd;
  if (z->fail) {
    if (z->fail->iday == iday && z->fail->shft == shft) {
      num_fail++;
      i_been_prob += 16;
    }
  } else {
    if (shft->schd->g > Imposible && i_been_prob > 16) {
      i_been_prob -= 4;
    }
  }

}

//----------------------------------------------------------
//  An empl_day that could do this shft_day has become committed to
//  doing something else.

void  shft_day::note_busy (void)
{
  if (badge != Badge_Id) abort ("Sday-nb1");

  emps_can--;
  shft->det_imp();

}

//----------------------------------------------------------
//  An empl_day is now in a position to work this shft_day
//  when the emp gets to this day.

void  shft_day::note_could (void)
{
  if (badge != Badge_Id) abort ("Sday-na1");
  emps_could++;
  shft->det_imp();
}

//----------------------------------------------------------
//  An empl_day is now in a position to work this shft_day now.

void  shft_day::note_can (void)
{
  if (badge != Badge_Id) abort ("Sday-na1");

  emps_could--;
  emps_can++;
  shft->det_imp();

}

//----------------------------------------------------------
//  I have little hope for improvement; I must change the way I look at
//  things so as to try different things.

void  shft_day::broaden_horizons (void)
{
  if (badge != Badge_Id) abort ("Sday-bh1");

  i_been_prob /= 2;

}

//---------------------------------------------------------
//  Check the data-structure integrity.

void  shft_day::check_integrity (void)
{
  if (badge != Badge_Id) abort ("Sday-Badge");

  if (emps_req < 0 || emps_req > shft->schd->num_emp) abort ("Sday-1");
  if (iday < Week_Start) {
    if (emps_have != emps_req) abort ("Sday-6");
    if ( ! done ) abort ("Sday-2");
  } else {
    if (emps_have) abort ("Sday-3");
//    if (done) abort ("Sday-5");
  }

}

//----------------------------------------------------------
