//  Brian Marshall      bmarshal@agt.net
//......................................

//  If a schedule file does not exist for the passed week_num, read
//  the Department and Employee files and write an empty new one.


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define  Null  0

void   write_header (FILE *sfp, int week_num, char *title,
                               int i_dow_start);
void   write_body (FILE *sfp);
char  *day_of_week (int i_dow_start, int iday);
int    get_name (char *rec, char *name);
int    read_dept (char *title, int *i_dow_start);
int    get_dept (char *rec, char *title, int *i_dow_start);

//----------------------------------------------------------
//  wf_new  main module.

void  main (int argc, char* argv[])
{

  char  title[61];
  int   week_num, i_dow_start;

  if (argc != 2) {
    printf ("\nUsage:  wf_new <week-num>\n\n");
    exit (1);
  }

  week_num = 0;
  week_num = atoi(argv[1]);
  if (week_num < 1 || week_num > 999) {
    printf ("\nWeek-Number must be between 1 and 999\n");
    exit (1);
  }

  char  sfilnam[201];               // The schedule file name.
  sprintf (sfilnam, "SCHD_%3.3d.WF", week_num);

  FILE *sfp = fopen (sfilnam, "r");
  if (sfp) {    // File does exist...
    char rec[101], *p;
    fgets (rec, 100, sfp);
    rec[99] = '\0';
    p = rec;
    while (*p == ' ' || *p == '|') p++;
    if (strncmp(p,"EMPTY",5) == 0 ||
        strncmp(p,"Empty",5) == 0 ||
        strncmp(p,"empty",5) == 0) {
      fclose (sfp);
      sfp = Null;  // User wants it emptied anyway.
    }
  }

  if (sfp) {                          // Exit if schedule already exists.
    fclose (sfp);
    exit(0);
  }

  if ( ! read_dept (title, &i_dow_start) ) {
    printf ("\nCould not find file DEPT.WF\n");
    printf ("Could not create new schedule file.\n");
    exit(1);
  }

  sfp = fopen (sfilnam, "w");
  if (!sfp) {
    printf ("Error creating schedule file. Abort\n");
    exit(1);
  }

  write_header (sfp, week_num, title, i_dow_start);

  write_body (sfp);

  fclose (sfp);

}

//----------------------------------------------------------
//  Write the schedule header.

void  write_header (FILE *sfp, int week_num, char *title,
                               int i_dow_start)
{
  char  rec[101], buf[31];

  strcpy (rec, "!  WorkForce Staff Schedule");
  strcat (rec, "                            Week: ");
  sprintf (buf, "%3.3d", week_num);
  strcat (rec, buf);
  fputs (rec, sfp);

  strcpy (rec, "\n\n!          ");
  strcat (rec, title);
  strcat (rec, "\n\n");
  fputs (rec, sfp);

  strcpy (rec, "!     Employee         |");
  int iday;
  for (iday=0; iday<7; iday++) {
    strcat (rec, "  ");
    strcat (rec, day_of_week(i_dow_start, iday));
    strcat (rec, " |");
  }
  strcat (rec, "\n");
  fputs (rec, sfp);

}

//----------------------------------------------------------
//  Write the schedule matrix.

void  write_body (FILE *sfp)
{
  char   rec[101], name[21];
  int    iday;

  FILE *efp = fopen ("EMP.WF", "r");            // Open Employee file.
  if (!efp) {
    printf ("Could not open file EMP.WF   Abort.\n");
    exit (1);
  }

  while ( ! feof(efp) ) {

    fgets (rec, 100, efp);
    if (ferror(efp)) {
      printf ("Error reading file EMP.WF   Abort.\n");
      exit (1);
    }

    if ( ! get_name (rec, name) ) continue;

    sprintf (rec, "   %-20s|", name);

    for (iday=0; iday<7; iday++) {
      strcat (rec, "      |");
    }

    strcat (rec, "\n");
    fputs (rec, sfp);

  }    // End of loop through records in emp file.

  fclose (efp);

}

//----------------------------------------------------------
//  Return the day of week abbreviation for day iday in the schedule.

char*  day_of_week (int i_dow_start, int iday)
{
  static char *dowa[7] = {"Mon", "Tue", "Wed", "Thu",
                          "Fri", "Sat", "Sun"};

  int i = i_dow_start + iday;
  while (i > 6) i -= 7;
  if (i < 0) i = 0;

  return dowa[i];

}

//----------------------------------------------------------
//  Get the name from the passed employee record.
//  Return true on success or false if rec is not a data record.

int  get_name (char *rec, char *name)
{
  char *p = rec;

  while (*p != '|') {                          // Skip emp-num column.
    if (*p == '!' || *p == '\0') return 0;
    p++;
  }
  p++;  // Skip the bar.

  while (*p == ' ') p++;

  int i;
  for (i=0; i<20; i++) {
    if (*p == '\0') return 0;
    if (*p == '|') break;
    name[i] = *p;
    p++;
  }
  name[i] = '\0';

  if (name[0] == '\0') return 0;

  return 1;

}

//----------------------------------------------------------
//  Read the Department file to determine the title and the
//  index of the day of the week of the first day.
//  Return true on success, false if could not open file.

int  read_dept (char *title, int *i_dow_start)
{
  char   rec[101];

  FILE *dfp = fopen ("DEPT.WF", "r");
  if (!dfp) return 0;

  while ( ! feof(dfp) ) {

    fgets (rec, 100, dfp);
    if (ferror(dfp)) {
      printf ("Error reading file DEPT.WF  Abort.\n");
      exit (1);
    }

    if ( get_dept (rec, title, i_dow_start) ) {
      fclose (dfp);
      return 1;
    }

  }  // End of loop through records in department file.

  fclose (dfp);
  return 0;

}

//----------------------------------------------------------
//  Get the title and index of day-of-week of first day from the
//  department record.
//  Return true on success or false if rec is not the data record.

int  get_dept (char *rec, char *title, int *i_dow_start)
{
  int i;

  char *p = rec;

  char dow[4];
  if (*p == '!') return 0;
  while (*p == ' ' || *p == '|') p++;
  strncpy (dow, p, 3);
  dow[3] = '\0';
  for (i=0; i<3; i++) if (dow[i] >='a') dow[i] -= ('a' - 'A');

  if (!strcmp(dow,"MON") ) {
    *i_dow_start = 0;
  } else if (!strcmp(dow,"TUE") ) {
    *i_dow_start = 1;
  } else if (!strcmp(dow,"WED") ) {
    *i_dow_start = 2;
  } else if (!strcmp(dow,"THU") ) {
    *i_dow_start = 3;
  } else if (!strcmp(dow,"FRI") ) {
    *i_dow_start = 4;
  } else if (!strcmp(dow,"SAT") ) {
    *i_dow_start = 5;
  } else if (!strcmp(dow,"SUN") ) {
    *i_dow_start = 6;
  } else {
    return 0;
  }

  while (*p != '|') {          // Go to end of this column.
    if (*p == '\0') return 0;
    p++;
  }
  p++;

  while (*p != '|') {          // Skip num-try column.
    if (*p == '\0') return 0;
    p++;
  }
  p++;

  while (*p == ' ') p++;

  for (i=0; i<60; i++) {
    if (*p == '\0') return 0;
    if (*p == '|') break;
    title[i] = *p;
    p++;
  }
  title[i] = '\0';

  return 1;

}

//----------------------------------------------------------
