//  Brian Marshall      bmarshal@agt.net
//......................................

//  WorkForce Version 1.2   Main-Line.
//
//  Create a staff schedule for week number <week-num> (a positive integer).
//
//  Usage:  workforc <week-num>
//
//  The following files will be read:
//      dept.wf
//      shift.wf
//      emp.wf
//      avail.wf
//      timeoff.wf
//      schd_xxx.wf    (where xxx is <week-num> minus one)

//  The following files are written:
//      schd_nnn.wf    (where nnn is <week-num>)
//      log.wf

#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "workforc.h"

//----------------------------------------------------------
//  WorkForce Main-Line.
//  Return zero on success, one on fail.

void  main (int argc, char* argv[])
{

  int   week_num;

  if (argc != 2) {
    cout << "\nUsage:  WORKFORC <week-num>\n\n";
    wfexit(1);
  }

  if ( ! open_log() ) {
    cout << "\nError opening log file.  Abort.\n";
    wfexit(1);
  }

  time_t t = time(NULL);
  wlog3 ("\nWorkForce v1.2  -  Run Log       ",
         ctime(&t), "\n\n");

  set_echo_log (True);

  week_num = 0;
  week_num = atoi(argv[1]);
  if (week_num < 1 || week_num > 999) {
    wlog ("\nWeek-Number must be between 1 and 999\n");
    wfexit(1);
  }

  wlog3 ("\n         Week Number: ", argv[1], "\n");

  wlog ("\nSetting up the problem...\n");

  schedule the_problem;                      // Construct a schedule.
  schedule *z = &the_problem;                // Attach a pointer.

  z->problem_setup (week_num);

  wlog ("\nBeginning to look for a good solution...\n");

  if (z->find_good_schedule()) {          // If we find a solution...
    wlog ("\nA solution has been found.\n");
    z->write_bsf();                       //   write the best-so-far.
    set_echo_log (False);
    z->log_worst_bsf_fill();    // Describe worst fill in log.
    set_echo_log (True);
  } else {
    wlog ("\nCould not find acceptable schedule.\n");
  }

  wlog ("\nWorkForce run completed.\n");
  wfexit(1);

}

//----------------------------------------------------------
//  Print a message to cout the user and exit with the passed code.

void  wfexit (int code)
{
  cout << "\n\nWorkForce is done.\n\n";

  exit (code);
}

//----------------------------------------------------------
